"""
default_wiki.py

Default Wiki writing classes.

Copyright (C) 2003 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""


class HTMLWriter:

    head = \
    (
        u"<html>\n"
        u"<head>\n"
        u'  <meta http-equiv="Content-Type" '
        u'content="text/html; charset=utf-8">\n'
        u"  <title>%s</title>\n"
        u'  <style type="text/css">li.empty { list-style-type: none }</style>\n'
        u"</head>\n"
        u"\n"
        u"<body>\n"
    )
    
    tail = \
    (
        u"</body>\n"
        u"</html>\n"
    )
    
    signatures = \
    {
        (u"'", 2): ( u"i", ),
        (u"'", 3): ( u"b", ),
        (u"'", 5): ( u"b", u"i" ),
        (u"_", 2): ( u"u", ),
        (u"=", 1): ( u"h1", ),
        (u"=", 2): ( u"h2", ),
        (u"=", 3): ( u"h3", ),
        (u"=", 4): ( u"h4", ),
        (u"=", 5): ( u"h5", ),
        (u"=", 6): ( u"h6", ),
        (u"`", 1): ( u"tt", ),
        (u"{", 3): ( u"tt", ),
        (u"}", 3): ( u"/tt", )
    }
    
    anchor = (u"[", u"]")
    preformatted = (u"{{{", u"}}}")
    image_formats = (u".png", u".jpg", u".gif")
    
    list_elements = \
    {
        "default":  {"begin": u"<ul>", "item": u"<li>", "end": u"</ul>"},
        u"*":       {"begin": u"<ul>", "item": u"<li>", "end": u"</ul>"},
        u"#":       { "begin": u'<ul>', "item": u'<li class="empty">',
                      "end": u"</ul>", "handler": "write_number" },
        u"@":       { "begin": u'<ul>', "item": u'<li class="empty">',
                      "end": u"</ul>", "handler": "write_alpha" },
        u"$":       { "begin": u'<ul>', "item": u'<li class="empty">',
                      "end": u"</ul>", "handler": "write_Alpha" }
    }
    
    def __init__(self):
    
        # Generate the list of special characters from the dictionary of
        # signatures.
        
        self.special = []
        
        for char, number in self.signatures.keys():
        
            self.special.append(char)
    
    def write(self, text, url):
    
        self.html = [self.head % url]
        
        self.text = text.split(u"\n")
        
        # Find the index of the next line to be examined and the body text
        # which was created.
        index, body = self.write_text()
        
        # Add the body text to the list of HTML lines.
        self.html = self.html + body
        
        self.html.append(self.tail)
    
    def output(self, export = 0):
    
        if export == 1:
        
            # Modify URLs given in anchors to allow pages to be exported.
            self.export_urls()
        
        return u"\n".join(self.html).encode("utf_8")
    
    def write_text(self, index = 0, depth = 0, indices = None):
    
        if indices is None: indices = []
        
        html = []
        
        preformatted = 0
        
        while index < len(self.text):
        
            line = self.text[index]
            
            if preformatted == 0 and depth == 0 and line.strip() == u"":
            
                html.append(u"<p>")
                index = index + 1
                continue
            
            # Determine the indent of the line.
            indent, rest = self.read_indent(line, u" ")
            
            if line.find(self.preformatted[0]) == 0:
            
                # Entering preformatted text.
                preformatted = 1
                html.append(u"<pre>")
                html.append(self.html_repr(rest[len(self.preformatted[0]):]))
                index = index + 1
            
            elif line.find(self.preformatted[1]) == 0:
            
                # Exiting preformatted text.
                preformatted = 0
                html.append(u"</pre>")
                html.append(self.write_line(rest[len(self.preformatted[1]):]))
                index = index + 1
            
            elif preformatted == 1:
            
                html.append(self.html_repr(line))
                index = index + 1
            
            elif rest[:1] in self.list_elements.keys():
            
                # This is a list element.
                
                # Read the format of the items in the list.
                
                format_str = self.read_list_format(rest)
                
                # Find the depth of the list nesting and the most recently
                # added type of list.
                
                new_depth, element = self.read_list_depth(format_str)
                
                if new_depth > depth:
                
                    # A greater depth of list elements.
                    
                    # For numeric lists, increase the number of indices
                    # used.
                    
                    # Increase the indentation counter but examine this
                    # line again at the new indentation.
                    
                    html.append(self.list_elements[element]["begin"])
                    
                    index, body = self.write_text(
                        index, new_depth,
                        indices[:] + (new_depth - depth - 1) * [1] + [0]
                        )
                    
                    html = html + body
                    
                    html.append(self.list_elements[element]["end"])
                
                elif new_depth < depth:
                
                    # A lesser depth of list elements.
                    
                    # Return the HTML found at this level to the caller.
                    
                    return index, html
                
                else:
                
                    # Write the list element according to whether the
                    # item is enumerated or not.
                    
                    if self.enumerated(format_str, len(indices)):
                    
                        # Increase the index corresponding to the current
                        # depth of nesting in the list.
                        
                        indices[-1] = indices[-1] + 1
                        
                        item_str = self.write_list_format(
                            indices[:], format_str
                            )
                        
                        item = self.list_elements[element]["item"] + \
                            self.html_repr(item_str)
                    
                    else:
                    
                        item = self.list_elements["default"]["item"]
                    
                    # Continue processing this line after the format string.
                    html.append(item)
                    html.append(self.write_line(rest[len(format_str):]))
                    
                    index = index + 1
            
            else:
            
                if depth > 0:
                
                    return index, html
                
                html.append(self.write_line(rest))
                index = index + 1
        
        return index, html
    
    def read_indent(self, line, indent_char):
    
        i = 0
        
        for c in line:
        
            if c == indent_char:
                i = i + 1
            else:
                break
        
        return i, line[i:]
    
    def write_line(self, line):
    
        text = []
        
        used = []
        
        i = 0
        
        while i < len(line):
        
            if line[i] in self.special:
            
                chars, rest = self.read_indent(line[i:], line[i])
                
                try:
                
                    tagnames = list(self.signatures[(line[i], chars)])
                    
                    # Remove any existing effects.
                    while len(used) > 0:
                    
                        if used[-1] in tagnames:
                        
                            # The most recently applied effect is listed
                            # in the attributes of this effect.
                            
                            # Write an end tag to indicate that the effect
                            # ends here.
                            
                            text.append(u"</"+used[-1]+">")
                            
                            # Remove the effect from both lists.
                            tagnames.remove(used[-1])
                            used.pop()
                        
                        else:
                        
                            # We cannot remove any more effects, although
                            # the list of used effects is not empty.
                            break
                    
                    # Apply any remaining new effects.
                    
                    for tagname in tagnames:
                    
                        # Add a tag to the text.
                        text.append(u"<"+tagname+">")
                        
                        # Add the tag name to the list of used effects
                        # unless it is a closing tag.
                        if tagname[0] != u"/":
                        
                            used.append(tagname)
                    
                    i = i + chars
                
                except KeyError:
                
                    text.append(line[i])
                    i = i + 1
            
            elif line[i] == self.anchor[0]:
            
                end_at = line.find(self.anchor[1], i+1)
                
                # Look for the end of the anchor, using the end of the line
                # if it cannot be found.
                
                if end_at == -1: end_at = len(line)
                
                # The contents of the anchor lie between the delimiters.
                # We remove any leading spaces.
                contents = line[i+1:end_at].lstrip()
                
                space_at = contents.find(u" ")
                
                if space_at == -1:
                
                    # No space was found so both the URL and the text
                    # are defined to be the same.
                    url = a_text = contents[1:]
                
                else:
                
                    # Separate the URL from the contents and remove any
                    # leading spaces from the text.
                    url = contents[:space_at]
                    a_text = contents[space_at:].lstrip()
                
                if url[:1] == u"#":
                
                    # Add the anchor tags with the text following after.
                    text.append(u'<a name="'+url[1:]+u'"></a>'+a_text)
                
                elif filter(lambda x: url.rfind(x) != -1, self.image_formats):
                
                    # This anchor refers to an image.
                    if url == a_text:
                    
                        text.append(
                            u'<a href="'+url+u'"><img src="'+url+u'"></a>'
                            )
                    
                    else:
                    
                        text.append(
                            u'<a href="'+url+u'"><img src="'+url+u'" '
                            u'alt="'+a_text+u'"></a>'
                            )
                
                else:
                
                    # Add the anchor tags and the text between them.
                    text.append(u'<a href="'+url+u'">'+a_text+u'</a>')
                
                # The index into the line is either the character after the
                # end of the anchor or the end of the line, depending on
                # whether an ending delimiter was found.
                i = min(end_at + 1, len(line))
            
            elif line[i] == u"\\":
            
                # The following character is written as it is but in an
                # appropriate form for HTML.
                
                following = line[i+1:i+2]
                text.append(self.html_repr(following))
                
                i = i + 2
            
            else:
            
                text.append(line[i])
                i = i + 1
        
        return u"".join(text)
    
    def html_repr(self, s):
    
        new = []
        
        for c in unicode(s):
        
            if c == u"&":
                new.append(u"&amp;")
            elif c == u"<":
                new.append(u"&lt;")
            elif c == u">":
                new.append(u"&gt;")
            else:
                new.append(c)
        
        return u"".join(new)
    
    def read_list_format(self, text):
    
        at = text.find(u" ")
        
        if at == -1:
        
            return text
        
        else:
        
            return text[:at]
    
    def read_list_depth(self, text):
    
        depth = 0
        item = u""
        
        for c in text:
        
            if c in self.list_elements.keys():
            
                depth = depth + 1
                item = c
        
        return depth, item
    
    def enumerated(self, format, n_indices):
    
        number = 0
        
        for item in self.list_elements.keys():
        
            if self.list_elements[item].has_key("handler"):
            
                # If the type of list represented by this item character
                # is an enumerated list then check whether the item
                # character is used in this format string.
                number = number + format.count(item)
        
        return number == n_indices
    
    def write_list_format(self, indices, format):
    
        output = []
        
        for c in format:
        
            # Examine each character in the format string.
            
            if c in self.list_elements.keys():
            
                # The character corresponds to a list item specifier.
                
                # Find a handler to convert the list index at this level
                # into a string.
                handler = self.list_elements[c].get("handler", None)
                
                if handler and hasattr(self, handler) and len(indices) > 0:
                
                    # A handler was available so call it, append the
                    # string to the output list and remove the index for
                    # this level.
                    
                    fn = getattr(self, handler)
                    output.append(fn(indices[0]))
                    indices.pop(0)
                
                else:
                
                    # There wasn't a handler specified, it didn't exist or
                    # there are no more indices to convert.
                    pass
            
            else:
            
                # Normal text.
                output.append(c)
        
        if len(indices) > 0:
        
            # Not all indices were converted.
            return None
        
        else:
        
            # Return the generated item string.
            return u"".join(output)
    
    def write_number(self, value):
    
        return u"%i" % value
    
    def write_alpha(self, value):
    
        if 0 < value < 27:
        
            return unicode(chr(96 + value))
        
        else:
        
            return u"?"
    
    def write_Alpha(self, value):
    
        return self.write_alpha(value).upper()
    
    def export_urls(self):
    
        for i in range(0, len(self.html)):
        
            item = self.html[i]
            
            if item.find(u'<a href="') != -1:
            
                start = 0
                new_item = u""
                
                while start < len(item):
                
                    anchor_at = item.find(u'<a href="', start)
                    
                    if anchor_at == -1:
                    
                        new_item = new_item + item[start:]
                        break
                    
                    else:
                    
                        # Add any preceding text to the new item.
                        new_item = new_item + \
                            item[start:anchor_at + len(u'<a href="')]
                        
                        # Extract the URL from the anchor tag.
                        anchor_end = item.find(
                            u'">', anchor_at + len(u'<a href="')
                            )
                        
                        if anchor_end == -1: anchor_end = len(item)
                        
                        url = item[anchor_at + len(u'<a href="'):anchor_end]
                        
                        # Determine whether the referenced item is to be found
                        # relative to the current page.
                        
                        if url.find(u":") == -1:
                        
                            # Assume it is a relative link.
                            
                            # Look for a suffix.
                            suffix_at = url.rfind(u".")
                            
                            if suffix_at == -1 or \
                                url[suffix_at:] not in self.image_formats:
                            
                                # The file either has no suffix or the suffix
                                # does not belong to that of a recognised image
                                # type.
                                
                                # Give the reference an ".html" suffix.
                                url = url + u".html"
                        
                        # Add the URL to the new item.
                        new_item = new_item + url + u'">'
                        
                        start = anchor_end + len(u'">')
                
                self.html[i] = new_item
                
