"""
wiki.py

Wiki writing classes.

Copyright (C) 2003 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""


class Template:

    text = \
u"""<html>
<head>
  <meta http-equiv="Content-Type" content="text/html; charset=%(charset)s">
  <title>%(title)s</title>
  <style type="text/css">
%(stylesheet)s
  
   </style>
</head>

<body>"""

    def head(self, stylesheet, title, charset = u"utf_8", export = 0):
    
        return self.text % {
            u"charset": charset,
            u"stylesheet": stylesheet.text,
            u"title": title
            }
    
    def tail(self):
    
        return u"</body>\n</html>\n"


class Stylesheet:

    text = \
u"""
    BODY
    {
        background-color: white;
        text-align: justify
    }
    
    LI.empty { list-style-type: none }
    
    PRE
    {
        background-color: #e0e0e0;
        color: black;
        font-size: 1.0em;
        padding: 0.4em;
        border: thin dotted black
    }
"""


class FormatElement:

    def __init__(self, text):
    
        self.text = u"<%s>" % text
        
        self.attributes = \
        {
            u"id": None,
            u"class": None
        }
    
    def replace_previous(self, elements, of_class = None):
    
        """replace_previous(self, elements, of_class = None)
        
        When supplied with a list or other mutable sequence object,
        this element replaces any continuous sequence of elements
        of the same class residing at the end of the list.
        
        The elements list is modified so that these preceding elements
        are removed and replaced with a single element which incorporates
        this element.
        """
        
        self.remove_previous(elements, of_class)
        
        # Insert a single element.
        elements.append(self)
    
    def remove_previous(self, elements, of_class = None):
    
        """remove_previous(self, elements, of_class = None)
        
        When supplied with a list or other mutable sequence object,
        this element removes any continuous sequence of elements
        of the same class residing at the end of the list.
        
        The elements list is modified so that these preceding elements
        are removed.
        """
        
        if of_class is None:
        
            of_class = self.__class__
        
        i = len(elements)
        
        while i > 0:
        
            if not isinstance(elements[i-1], of_class):
            
                break
            
            else:
            
                i = i - 1
        
        # Remove the elements from the end of the list.
        elements[:] = elements[:i]
    
    def find_previous(self, elements, of_class = None, all = 0):
    
        """find_previous(self, elements, of_class = None, all = 0)
        
        When supplied with a list or other mutable sequence object,
        this element finds previous elements of the same class (or
        of the class specified) in the list, returning a list of
        those elements in reverse order.
        If all is set to a non-zero value, all previous elements of
        the required class are returned.
        """
        
        if of_class is None:
        
            of_class = self.__class__
        
        i = len(elements)
        
        found = []
        
        while i > 0:
        
            if isinstance(elements[i-1], of_class):
            
                found.append(elements[i-1])
            
            i = i - 1
        
        # Return matching elements in a list.
        return found
    
    def merge_whitespace(self, elements):
    
        """merge_whitespace(self, elements)
        
        When supplied with a list or other mutable sequence object,
        this element is merged with any continuous sequence of elements
        of the class specified residing at the end of the list.
        
        The elements list is modified so that these preceding elements
        are removed and replaced with a single element which incorporates
        this element.
        """
        
        i = len(elements)
        
        # Record the total amount of whitespace found.
        amount = 0
        
        while i > 0:
        
            element = elements[i-1]
            
            if not isinstance(element, Text):
            
                break
            
            elif element.text.strip() != u"":
            
                # Not all characters are whitespace.
                
                left_text = element.text.rstrip()
                left_length = len(left_text)
                
                if left_text != element.text:
                
                    # Trailing whitespace exists.
                    
                    # Increase the amount of whitespace by the trailing
                    # whitespace for this mixed element.
                    amount = amount + len(element.text) - left_length
                    
                    # Remove the trailing whitespace characters.
                    elements[i-1].text = element.text[:left_length]
                
                # Leave the loop since we have found non-whitespace
                # characters.
                break
            
            else:
            
                i = i - 1
                amount = amount + len(elements[i].text)
        
        # Only insert a replacement whitespace element if whitespace was
        # found.
        
        if amount > 0:
        
            # Remove the whitespace elements from the end of the list.
            elements[:] = elements[:i]
            
            # Insert a single whitespace element.
            elements.append( Text(u" "*amount, whitespace = 1) )



class FormatBegin(FormatElement):

    def __init__(self, text, attributes = None):
    
        self.name = text
        
        self.attributes = \
        {
            u"id": None,
            u"class": None
        }
        
        self.update_attributes(attributes)
    
    def update_attributes(self, pairs = None):
    
        if pairs:
        
            for key, value in pairs:
            
                self.attributes[key] = value
        
        attrs = filter(
            lambda pair: pair[1], self.attributes.items()
            )
        
        if attrs != []:
        
            attrs = u" ".join(
                map(lambda pair: u'%s="%s"' % pair, attrs)
                )
        
            self.text = u"<%s %s>" % (self.name, attrs)
        
        else:
        
            self.text = u"<%s>" % self.name


class FormatEnd(FormatElement):

    def __init__(self, text, attributes = None):
    
        self.text = u"</%s>" % text
        
        self.attributes = \
        {
            u"id": None,
            u"class": None
        }
        
        self.update_attributes(attributes)
    
    def update_attributes(self, pairs = None):
    
        if pairs:
        
            for key, value in pairs:
            
                self.attributes[key] = value


class Text(FormatElement):

    def __init__(self, text, whitespace = 0):
    
        self.text = text
        self.whitespace = whitespace

class EffectBegin(FormatBegin):

    pass

class EffectEnd(FormatEnd):

    pass

class Break(FormatElement):

    def __init__(self, text):
    
        self.text = text

class LineBreak(Break):

    def __init__(self, text = u"\n"):
    
        Break.__init__(self, text)

class ParagraphBreak(FormatBegin):

    def process(self, elements, indent = 0):
    
        if elements != [] and isinstance(elements[-1], Text):
        
            # Merge preceding whitespace into a single element.
            self.merge_whitespace(elements)
            
            # Remove any preceding whitespace.
            if elements[-1].whitespace == 1:
            
                elements.pop()
        
        # If the previous element is not a heading, paragraph break, or a
        # named anchor then add this element to the list of elements.
        
        if elements == []:
        
            elements.append(self)
        
        elif isinstance(elements[-1], HeadingEnd):
        
            pass
        
        elif isinstance(elements[-1], ParagraphBreak):
        
            pass
        
        elif isinstance(elements[-1], AnchorName):
        
            pass
        
        else:
        
            elements.append(self)
        
        # Return the indentation level unchanged.
        return indent

class BlockBegin(FormatBegin):

    def __repr__(self):
    
        return "<BlockBegin %s>" % repr(self.attributes.items())
    
    def process(self, elements, indent = 0):
    
        elements.append(self)
        return indent + 2

class BlockEnd(FormatEnd):

    def __repr__(self):
    
        return "<BlockEnd %s>" % repr(self.attributes.items())
    
    def process(self, elements, indent = 0):
    
        elements.append(self)
        return indent - 2

class Class(FormatElement):

    def __init__(self, value):
    
        self.value = value
    
    def process(self, elements, indent = 0):
    
        # Unlike other formatting elements, Class elements never
        # include themselves in the page body, but instead modify
        # previous starting elements.
        
        # Look for a relevant starting element.
        found = filter(
            lambda x: isinstance(x, FormatBegin) and \
                x.attributes[u"class"] == None,
            elements
            )
        
        if found != []:
        
            # Modify the element found.
            found[-1].update_attributes([(u"class", self.value)])
        
        return indent

class Identity(FormatElement):

    def __init__(self, value):
    
        self.value = value
    
    def process(self, elements, indent = 0):
    
        # Unlike other formatting elements, Identity elements never
        # include themselves in the page body, but instead modify
        # previous elements.
        
        # Look for a corresponding starting identity.
        found = filter(
            lambda x: isinstance(x, BlockBegin) and \
                x.attributes[u"id"] == self.value,
            elements
            )
        
        if found != [] and self.value:
        
            ending = filter(
                lambda x: isinstance(x, BlockEnd) and \
                    x.attributes[u"id"] == self.value,
                elements
                )
            
            if ending == []:
            
                # Ensure that the list is not preceded by a paragraph break.
                self.remove_previous(elements, of_class = ParagraphBreak)
                
                # Insert a BlockEnd element.
                elements.append( BlockEnd( u"div", ((u"id", self.value),) ) )
        
        elif found != []:
        
            # Anonymous block elements
            
            # Ensure that the list is not preceded by a paragraph break.
            self.remove_previous(elements, of_class = ParagraphBreak)
            
            # Insert a BlockEnd element.
            elements.append( BlockEnd( u"div", ((u"id", self.value),) ) )
            
            # Insert a block starting element with this ID.
            elements.append(
                BlockBegin( u"div", attributes = ((u"id", self.value),) )
                )
        
        else:
        
            # Ensure that the new block is not preceded by a paragraph break.
            self.remove_previous(elements, of_class = ParagraphBreak)
            
            # Insert a block starting element with this ID.
            elements.append(
                BlockBegin( u"div", attributes = ((u"id", self.value),) )
                )
        
        return indent

class ListBegin(FormatBegin):

    def process(self, elements, indent = 0):
    
        if elements != [] and isinstance(elements[-1], Text):
        
            # Merge preceding whitespace into a single element.
            self.merge_whitespace(elements)
            
            # Remove any preceding whitespace.
            if elements[-1].whitespace == 1:
            
                elements.pop()
        
        # Ensure that the list is not preceded by a paragraph break.
        self.remove_previous(elements, of_class = ParagraphBreak)
        
        if indent > 0:
        
            # Indent this element appropriately.
            elements.append( Text(u" " * indent) )
        
        elements.append(self)
        elements.append( LineBreak() )
        
        return indent + 2

class ListEnd(FormatEnd):

    def process(self, elements, indent = 0):
    
        if elements != [] and isinstance(elements[-1], Text):
        
            # Merge preceding whitespace into a single element.
            self.merge_whitespace(elements)
            
            # Remove any preceding whitespace.
            if elements[-1].whitespace == 1:
            
                elements.pop()
        
            # If the preceding element is not a line break then insert one.
            if not isinstance(elements[-1], LineBreak):
            
                elements.append( LineBreak() )
        
        indent = max(indent - 2, 0)
        
        if indent > 0:
        
            # Indent this element appropriately.
            elements.append( Text(u" " * indent) )
        
        elements.append(self)
        elements.append( LineBreak() )
        
        return indent

class ListItem(FormatBegin):

    def process(self, elements, indent = 0):
    
        # Work back to the previous structural element or non-whitespace
        # text.
        
        while elements != []:
        
            element = elements[-1]
            
            if isinstance(element, Text):
            
                # Merge preceding whitespace into a single element.
                self.merge_whitespace(elements)
                
                # Remove any preceding whitespace.
                if elements[-1].whitespace == 1:
                
                    elements.pop()
                
                else:
                
                    # We cannot continue further.
                    break
            
            elif isinstance(element, LineBreak):
            
                # Discard this element.
                elements.pop()
            
            else:
            
                # Do not continue further.
                break
        
        # Insert a line break.
        elements.append( LineBreak() )
        
        if indent > 0:
        
            # Indent this element appropriately.
            elements.append( Text(u" " * indent) )
        
        # Append this element to the list of elements.
        elements.append(self)
        
        return indent

class HeadingBegin(FormatBegin):

    def __init__(self, text, level):
    
        FormatBegin.__init__(self, text)
        self.level = level
    
    def process(self, elements, indent = 0):
    
        # Ensure that the heading is not preceded by a paragraph break.
        self.remove_previous(elements, of_class = ParagraphBreak)
        
        # For readability, ensure that this element, if not the first
        # in the document, is preceded by a line break.
        
        self.remove_previous(elements, of_class = LineBreak)
        
#        # Find any previous headings to determine whether to end the
#        # current block.
#        headings = self.find_previous(elements)
#        
#        if headings != []:
#        
#            # A previous heading was found.
#            
#            # If the previous heading has an equal or higher level value
#            # then end the block.
#            if headings[0].level <= self.level:
#            
#                elements.append( BlockEnd(u"div") )
        
        elements.append( LineBreak() )
        
        # Append this element to the list of elements.
        elements.append(self)
        
        # Return the indentation level unchanged.
        return indent

class HeadingEnd(FormatEnd):

    def __init__(self, text, level):
    
        FormatEnd.__init__(self, text)
        self.level = level
    
    def process(self, elements, indent = 0):
    
        elements.append(self)
        elements.append( LineBreak() )
        
        # Return the indentation level unchanged.
        return indent

class PreformattedBegin(FormatBegin):

    def process(self, elements, indent = 0):
    
        # Ensure that the heading is not preceded by a paragraph break.
        self.remove_previous(elements, of_class = ParagraphBreak)
        
        # For readability, ensure that this element, if not the first
        # in the document, is preceded by two line breaks.
        
        self.remove_previous(elements, of_class = LineBreak)
        
        elements.append( LineBreak() )
        elements.append( LineBreak() )
        
        # Append this element to the list of elements.
        elements.append(self)
        
        # Return the indentation level unchanged.
        return indent

class PreformattedEnd(FormatEnd):

    def process(self, elements, indent = 0):
    
        elements.append(self)
        elements.append( LineBreak() )
        
        # Return the indentation level unchanged.
        return indent

class AnchorName(FormatElement):

    def process(self, elements, indent = 0):
    
        # Ensure that the heading is not preceded by a paragraph break.
        self.remove_previous(elements, of_class = ParagraphBreak)
        
        elements.append(self)
        
        # Return the indentation level unchanged.
        return indent

class AnchorBegin(FormatBegin):

    def __init__(self, text, href):
    
        FormatBegin.__init__(self, text, ((u"href", href),))

class AnchorEnd(FormatEnd):

    pass

class Image(FormatBegin):

    image_formats = (u".png", u".jpg", u".gif")


class HTMLWriter:

    # Define class variables containing the formatting classes to be used
    # by the formatter. These may be overridden by subclasses.
    
    EffectBegin = EffectBegin
    EffectEnd = EffectEnd
    Text = Text
    LineBreak = LineBreak
    ParagraphBreak = ParagraphBreak
    BlockBegin = BlockBegin
    BlockEnd = BlockEnd
    Class = Class
    Identity = Identity
    ListBegin = ListBegin
    ListEnd = ListEnd
    ListItem = ListItem
    HeadingBegin = HeadingBegin
    HeadingEnd = HeadingEnd
    PreformattedBegin = PreformattedBegin
    PreformattedEnd = PreformattedEnd
    AnchorName = AnchorName
    AnchorBegin = AnchorBegin
    AnchorEnd = AnchorEnd
    Image = Image
    Template = Template
    Stylesheet = Stylesheet
    
    # Input signatures
    
    signatures = \
    {
        (u"'", 2): ( u"i", ),
        (u"'", 3): ( u"b", ),
        (u"'", 5): ( u"b", u"i" ),
        (u"_", 2): ( u"u", ),
        (u"`", 1): ( u"tt", )
    }
    
    headings = \
    {
        (u"=", 1): u"h1",
        (u"=", 2): u"h2",
        (u"=", 3): u"h3",
        (u"=", 4): u"h4",
        (u"=", 5): u"h5",
        (u"=", 6): u"h6"
    }
    
    anchor = (u"[", u"]")
    preformatted = (u"{{{", u"}}}")
    id_label = u"|"
    class_label = u"^"
    
    def __init__(self):
    
        # Generate the list of special characters from the dictionary of
        # signatures.
        
        self.special = []
        
        for char, number in self.signatures.keys():
        
            self.special.append(char)
        
        # Define the document head.
        
        self.template = self.Template()
        self.stylesheet = self.Stylesheet()
        
        # Output elements
        
        self.list_elements = \
        {
            "default":
            {
                "begin":    (self.ListBegin, u"ul"),
                "item":     (self.ListItem, u"li"),
                "end":      (self.ListEnd, u"ul")
            },
            
            u"*":
            {
                "begin":    (self.ListBegin, u"ul"),
                "item":     (self.ListItem, u"li"),
                "end":      (self.ListEnd, u"ul")
            },
            
            u"#":
            {
                "begin":    (self.ListBegin, u"ul"),
                "item":     (self.ListItem, u'li class="empty"'),
                "end":      (self.ListEnd, u"ul"),
                "handler":  "write_number"
            },
            
            u"@":
            {
                "begin":    (self.ListBegin, u"ul"),
                "item":     (self.ListItem, u'li class="empty"'),
                "end":      (self.ListEnd, u"ul"),
                "handler":  "write_alpha"
            },
            
            u"$":
            {
                "begin":    (self.ListBegin, u"ul"),
                "item":     (self.ListItem, u'li class="empty"'),
                "end":      (self.ListEnd, u"ul"),
                "handler":  "write_Alpha"
            }
        }
        
        # Warning: these elements will be reused throughout the document,
        # so any changes to one instance will affect them all.
        
        self.line_break = self.LineBreak()
        self.anchor_end = self.AnchorEnd(u'a')
        self.in_anchor = 0
    
    def write(self, text, url):
    
        """write(self, text, url)
        
        Convert the text supplied from the URL specified into HTML with
        the UTF-8 character encoding. The result is stored internally
        until the output method is called.
        """
        
        self.text = text
        self.url = url
        
        # Set up an empty list for the generated body text.
        self.body = []
        
        # Initialise the character index into the text and the index of the
        # start of the current line.
        self.index = 0
        self.line_index = 0
        
        # Maintain a list of effects.
        self.effects = []
        
        # Maintain a list of list details.
        self.list_details = []
        self.list_counters = []
        
        # Maintain a flag indicating whether preformatted text is being
        # written.
        self.in_preformatted = 0
        
        self._write()
    
    def output(self, export = 0, charset = "utf_8"):
    
        """html = output(self, export = 0, charset = "utf_8")
        
        Return the document as HTML in a form suitable for rendering
        within a Wiki. If export is non-zero, links within the Wiki are
        converted to refer to HTML versions of documents; this is useful
        when exporting the contents of a Wiki for archiving purposes.
        The character set used to encode the text is set using the
        charset argument, defaulting to the UTF 8 encoding.
        """
        
        # When exporting, don't use the local URL as the document title.
        
        if export == 0:
        
            title = self.url
        
        else:
        
            at = self.url.rfind(u"/")
            
            if at == -1:
                title = self.url
            else:
                title = self.url[at+1:]
        
        html = [self.template.head(
            self.stylesheet, title, charset = charset, export = export
            )]
        
        if export == 1:
        
            # Modify URLs given in anchors to allow pages to be exported.
            self.export_urls()
        
        # Add the body text to the list of HTML lines.
        html.append(self.process_body())
        
        html.append(self.template.tail())
        
        if charset is not None:
        
            return u"\n".join(html).encode(charset)
        
        else:
        
            return u"\n".join(html)
    
    def _write(self):
    
        while self.index < len(self.text):
        
            # Read a character from the text.
            c = self.text[self.index]
            
            # Check whether the character is used to mark up some structure
            # in the text.
            fn = self.lookup.get( c, None )
            
            if fn:
            
                # Call the function which processes this markup.
                if fn(self, c): return
            
            else:
            
                # The character is just normal text.
                self.body.append( self.Text(self.html_repr(c)) )
                self.index = self.index + 1
    
    def _new_line(self, c):
    
        # Remove any existing effects.
        while len(self.effects) > 0:
        
            # The most recently applied effect is listed
            # in the attributes of this effect.
            
            # Write an end tag to indicate that the effect
            # ends here.
            
            self.body.append( self.EffectEnd(self.effects[-1]) )
            
            # Remove the effect from both lists.
            self.effects.pop()
        
        index = self.index
        line_index = self.line_index
        
        self.index = self.index + 1
        self.line_index = self.index
        
        if self.list_details != []:
        
            # If we are currently in a list then determine whether the new
            # line is sufficiently indented for list items.
            
            indent = self.read_indent()
            next = self.line_index + indent
            
            if self.text.find(self.list_details[-1]["character"], next) == next:
            
                # We have encountered another list item of the same type.
                # Let the list handler deal with this.
                pass
            
            elif indent >= self.list_details[-1]["indent"]:
            
                # The indent for the text (this includes other types of
                # list) equals or exceeds that of the list indent.
                
                # Let the list handler catch any new type of list.
                pass
            
            elif indent == 0:
            
                # Leave the list structure completely. We achieve this by
                # returning without moving the index onto the next character.
                
                self.index = index
                self.line_index = line_index
                return 1
            
            else:
            
                # Leave this level of the list structure but move onto
                # the next character.
                return 1
        
        elif index - line_index == 0 and not self.in_preformatted:
        
            # The newline character occurs immediately after the
            # previous one. Outside lists and other structures, this
            # means that a paragraph break is required.
            
            # Chains of paragraph breaks are removed when the document
            # body is later processed.
            
            self.body.append(self.ParagraphBreak(u"p"))
        
        else:
        
            # Include a line breaks for formatting purposes.
            
            # Chains of line breaks are removed when the document body
            # is later processed.
            
            self.body.append(self.line_break)
    
    def _write_effect(self, c):
    
        # Count the number of characters used to specify the effect.
        number, self.index = self.count_chars(self.index, c)
        
        try:
        
            tagnames = list(self.signatures[(c, number)])
            
            # Remove any existing effects.
            while len(self.effects) > 0:
            
                if self.effects[-1] in tagnames:
                
                    # The most recently applied effect is listed
                    # in the attributes of this effect.
                    
                    # Write an end tag to indicate that the effect
                    # ends here.
                    
                    self.body.append( self.EffectEnd(self.effects[-1]) )
                    
                    # Remove the effect from both lists.
                    tagnames.remove(self.effects[-1])
                    self.effects.pop()
                
                else:
                
                    # We cannot remove any more effects, although
                    # the list of used effects is not empty.
                    break
            
            # Apply any remaining new effects.
            
            for tagname in tagnames:
            
                # Add a tag to the text.
                self.body.append( self.EffectBegin(tagname) )
                
                # Add the tag name to the list of used effects
                # unless it is a closing tag.
                if tagname[0] != u"/":
                
                    self.effects.append(tagname)
        
        except KeyError:
        
            self.body.append( self.Text(self.html_repr(c * number)) )
    
    def _write_heading(self, c):
    
        # Count the number of characters used to specify the heading.
        number, self.index = self.count_chars(self.index, c)
        
        if hasattr(self, "heading_tag"):
        
            return 1
        
        try:
        
            tagname = self.headings[(c, number)]
        
        except KeyError:
        
            self.body.append( self.Text(self.html_repr(c * number)) )
            return
        
        # Write the heading to the list of elements used in the document
        # body.
        self.body.append( self.HeadingBegin(tagname, number) )
        self.heading_tag = tagname
        
        self._write()
        
        del self.heading_tag
        self.body.append( self.HeadingEnd(tagname, number) )
    
    def _write_id(self, c):
    
        # Read until the end of the label.
        self.index = self.index + 1
        
        label = []
        
        while self.index < len(self.text):
        
            c = self.text[self.index]
            self.index = self.index + 1
            
            if c == self.id_label:
                break
            
            label.append(c)
        
        label = u"".join(label).strip()
        
        if label != u"":
        
            self.body.append( self.Identity(label) )
    
    def _write_class(self, c):
    
        # Read until the end of the label.
        self.index = self.index + 1
        
        label = []
        
        while self.index < len(self.text):
        
            c = self.text[self.index]
            self.index = self.index + 1
            
            if c == self.class_label:
                break
            
            label.append(c)
        
        label = u"".join(label).strip()
        
        self.body.append( self.Class(label) )
    
    def _write_preformatted(self, c):
    
        # Count the number of characters used to specify the beginning of
        # preformatted text.
        number, self.index = self.count_chars(self.index, c)
        
        if number < len(self.preformatted[0]):
        
            self.body.append( self.Text(self.html_repr(c * number)) )
            return
        
        # Remove any preceding paragraph breaks.
        if self.body[-1:] == [self.ParagraphBreak(u"p")]:
        
            self.body.pop()
        
        self.body.append( self.PreformattedBegin(u"pre") )
        
        self.body.append(
            self.Text(self.html_repr(c) * (number - len(self.preformatted[0])))
            )
        
        end_c = self.preformatted[1][0]
        
        while self.index < len(self.text):
        
            c = self.text[self.index]
            
            if c == u"\\":
            
                self._escape_character(c)
            
            elif c == end_c:
            
                # Count the number of characters used to specify the end of
                # preformatted text.
                number, self.index = self.count_chars(self.index, end_c)
                
                if number != len(self.preformatted[1]):
                
                    self.body.append( self.Text(self.html_repr(end_c * number)) )
                
                else:
                
                    break
            else:
            
                self.body.append( self.Text(self.html_repr(c)) )
                self.index = self.index + 1
        
        self.body.append( self.PreformattedEnd(u"pre") )
    
    def _begin_code(self, c):
    
        # Count the number of characters used to specify the beginning of
        # preformatted text.
        number, self.index = self.count_chars(self.index, c)
        
        if self.in_preformatted or number < len(self.preformatted[0]) or \
            self.read_indent() > 0:
        
            self.body.append( self.Text(self.html_repr(c * number)) )
            return
        
        self.body.append( self.PreformattedBegin(u"pre") )
        self.in_preformatted = 1
        
        self._write()
        
        self.in_preformatted = 0
        self.body.append( self.PreformattedEnd(u"pre") )
    
    def _end_code(self, c):
    
        # Count the number of characters used to specify the end of
        # preformatted text.
        number, self.index = self.count_chars(self.index, c)
        
        if not self.in_preformatted or self.read_indent() > 0:
        
            # We are not currently in preformatted text.
            self.body.append( self.Text(self.html_repr(c) * number) )
            return
        
        elif number < len(self.preformatted[1]):
        
            self.body.append( self.Text(self.html_repr(c * number)) )
            return
        
        # We are in preformatted text. Return 1 to indicate that the
        # processing method should return.
        
        return 1
    
    def _write_list(self, c):
    
        # Determine whether the character occurs before any other
        # non-whitespace characters on a line.
        if not self.first_on_line():
        
            # Non-whitespace characters were found.
            self.body.append( self.Text(self.html_repr(c)) )
            self.index = self.index + 1
            return
        
        # Read the formatting string for the list item.
        format_str, new_index = self.read_list_format()
        
        # Keep the index of the first character in the formatting string.
        first_index = self.index
        
        # Find the depth from the format string.
        depth, element = self.find_list_depth(format_str)
        
        if self.list_details == []:
        
            # We will start a new list.
            
            # Position the index into the text to point to the following
            # text.
            self.index = new_index
        
        elif depth == self.list_details[-1]["depth"]:
        
            if format_str == self.list_details[-1]["format"]:
            
                # The format of the items at this level is unchanged.
                
                # Increment the counter for this level.
                self.list_counters[-1] = self.list_counters[-1] + 1
                
                # Construct the list item text.
                list_class, list_string = self.list_details[-1]["list item"]
                
                item_text = \
                [
                    list_class(list_string),
                    self.Text( self.html_repr(
                        self.write_list_format(format_str)
                        ) )
                ]
                
                # Include the list item markup in the body text.
                # This will be a subclass of FormatElement, so we don't
                # have to create a Text object to enclose it.
                
                self.body = self.body + item_text
                
                # Position the index into the text to point to the following
                # text.
                self.index = new_index
                return
            
            else:
            
                # Mixed format lists at the same level are not allowed.
                
                # Leave the list before starting a new one.
                return 1
        
        elif depth > self.list_details[-1]["depth"]:
        
            # Descend into a more deeply nested list.
            
            # Position the index into the text to point to the following
            # text.
            self.index = new_index
        
        else:
        
            # Ascend to a higher level of nesting or leave the list entirely.
            return 1
        
        # This is a list item from a deeper level.
        
        # Start a new list and read each item, continuing until text is
        # found which is not sufficiently indented, is part of a different
        # kind of list, or incorporates a different level of nesting.
        
        # Remove any preceding whitespace which was found on this line.
        #if self.body != []:
        #
        #    if self.body[-1].text == (u" " * (first_index - self.line_index)):
        #    
        #        self.body.pop()
        
        # Remove any preceding paragraph breaks.
        if self.body[-1:] == [self.ParagraphBreak(u"p")]:
        
            self.body.pop()
        
        list_item = self.list_elements[c]
        
        # Append the formatting element for the beginning of the list.
        list_class, list_string = list_item["begin"]
        
        self.body.append(list_class(list_string))
        self.list_details.append(
            {
                "character":    c,
                "indent":       max(first_index - self.line_index, 1),
                "format":       format_str,
                "depth":        depth,
                "list item":    list_item["item"]
            }
            )
        
        # If the depth increases by more than one level at a time
        # then insert a sufficient number of newly initialised counters
        # into the list in order to provide enough for the formatting
        # algorithm.
        
        old_depth = len(self.list_counters)
        
        self.list_counters = self.list_counters + (
            [1] * (depth - old_depth)
            )
        
        # Construct the first list item text.
        list_class, list_string = self.list_details[-1]["list item"]
        
        item_text = \
        [
            list_class(list_string),
            self.Text( self.html_repr(
                self.write_list_format(format_str)
                ) )
        ]
        
        # Include the list item formatting element in the body list.
        self.body = self.body + item_text
        
        self._write()
        
        self.list_details.pop()
        
        # Remove only the counters added before the descent.
        self.list_counters = self.list_counters[:old_depth]
        
        # Append the list end element to the body list.
        list_class, list_string = list_item["end"]
        
        self.body.append(list_class(list_string))
    
    def _escape_character(self, c):
    
        # Read the next character.
        character = self.text[self.index + 1:self.index + 2]
        
        self.body.append( self.Text(self.html_repr(character)) )
        
        self.index = self.index + 2
    
    def _begin_anchor(self, c):
    
        if self.in_anchor == 1:
        
            # We are already in an anchor. Instead we write the character
            # as given.
            self.body.append( self.Text(self.html_repr(c)) )
            self.index = self.index + 1
            return
        
        # Read the text following the starting character.
        i = self.index + 1
        
        while i < len(self.text):
        
            if self.text[i] == u" ":
                break
            elif self.text[i] == u"\n":
                break
            elif self.text[i] == self.anchor[1]:
                break
            else:
                i = i + 1
        
        url = self.text[self.index + 1:i]
        
        self.index = i + 1
        
        if self.text.find(self.anchor[1], i) == i:
        
            # Only a URL was specified.
            
            if url == u"":
            
                pass
            
            elif url[:1] != u"#":
            
                self._insert_simple_anchor(url)
            
            else:
            
                self.body.append( self.AnchorName(u'a name="%s"></a' % url[1:]) )
        
        elif url[-4:] in self.Image.image_formats:
        
            image = self.Image(u'img')
            image.update_attributes( ((u"src", url),) )
            
            self.body.append(image)
            self.in_anchor = 1
            
            self._write()
            
            self.in_anchor = 0
            
            item = len(self.body) - 1
            
            while self.body[item] != image:
            
                item = item - 1
            
            self.body, alt_text_items = self.body[:item+1], self.body[item+1:]
            
            alt_text = u"".join(
                map(lambda x: x.text, alt_text_items)
                )
            
            image.update_attributes( ((u"alt", alt_text),) )
        
        else:
        
            self.body.append( self.AnchorBegin(u"a", url) )
            self.in_anchor = 1
            
            self._write()
            
            self.in_anchor = 0
            self.body.append( self.anchor_end )
    
    def _insert_simple_anchor(self, url):
    
        if url[-4:] in self.Image.image_formats:
        
            self.body.append( self.Image(u'img', ((u"src", url),)) )
        
        else:
        
            self.body = self.body + \
            [
                self.AnchorBegin(u"a", url),
                self.Text(url),
                self.anchor_end
            ]
    
    def _end_anchor(self, c):
    
        if self.in_anchor == 1:
        
            self.index = self.index + 1
            return 1
        
        # We are not actually in an anchor. Instead we write the character
        # as given.
        self.body.append( self.Text(self.html_repr(c)) )
        self.index = self.index + 1
    
    # Create a lookup table for quick dispatch during the reading loop.
    lookup = \
    {
        u"'":   _write_effect,
        u"_":   _write_effect,
        u"=":   _write_heading,
        u"`":   _write_effect,
        
        preformatted[0][0]: _write_preformatted,
        #preformatted[1][0]: _end_code,
        
        u"*":   _write_list,
        u"@":   _write_list,
        u"$":   _write_list,
        u"#":   _write_list,
        #u"-":   _write_table,
        u"\n":  _new_line,
        u"\\":  _escape_character,
        
        id_label:   _write_id,
        class_label:   _write_class,
        
        anchor[0][0]:   _begin_anchor,
        anchor[1][0]:   _end_anchor
    }
    
    def count_chars(self, index, character):
    
        i = 0
        
        while index < len(self.text):
        
            if self.text[index] == character:
                i = i + 1
            else:
                break
            index = index + 1
        
        return i, index
    
    def read_indent(self):
    
        i = self.line_index
        
        while i < len(self.text):
        
            if self.text[i] == u" ":
                i = i + 1
            else:
                break
        
        # If there was only whitespace on the line then return an indent
        # of zero.
        #if i == len(self.text) or self.text[i] == u"\n":
        #
        #    return 0
        #
        #else:
        
        return i - self.line_index
    
    def first_on_line(self):
    
        if self.text[self.line_index:self.index].strip() != u"":
        
            return 0
        
        return 1
    
    def html_repr(self, s):
    
        new = []
        
        for c in unicode(s):
        
            if c == u"&":
                new.append(u"&amp;")
            elif c == u"<":
                new.append(u"&lt;")
            elif c == u">":
                new.append(u"&gt;")
            else:
                new.append(c)
        
        return u"".join(new)
    
    def read_list_format(self):
    
        i = self.index
        
        while i < len(self.text):
        
            if self.text[i] == u" ":
            
                # We want to skip this space.
                i = i + 1
                break
            
            elif self.text[i] == u"\n":
            
                # Don't skip any newlines.
                break
            
            else:
                i = i + 1
        
        return self.text[self.index:i], i
    
    def find_list_depth(self, text):
    
        depth = 0
        item = u""
        
        for c in text:
        
            if c in self.list_elements.keys():
            
                depth = depth + 1
                item = c
        
        return depth, item
    
    def enumerated(self, format, n_indices):
    
        number = 0
        
        for item in self.list_elements.keys():
        
            if self.list_elements[item].has_key("handler"):
            
                # If the type of list represented by this item character
                # is an enumerated list then check whether the item
                # character is used in this format string.
                number = number + format.count(item)
        
        return number == n_indices
    
    def write_list_format(self, format):
    
        output = []
        
        # Extract the list counters from the relevant instance variable.
        counters = self.list_counters[:]
        
        for c in format:
        
            # Examine each character in the format string.
            
            if c in self.list_elements.keys():
            
                # The character corresponds to a list item specifier.
                
                # Find a handler to convert the list index at this level
                # into a string.
                handler = self.list_elements[c].get("handler", None)
                
                if handler and hasattr(self, handler) and len(counters) > 0:
                
                    # A handler was available so call it, append the
                    # string to the output list and remove the index for
                    # this level.
                    
                    fn = getattr(self, handler)
                    output.append(fn(counters[0]))
                    counters.pop(0)
                
                else:
                
                    # There wasn't a handler specified, it didn't exist or
                    # there are no more indices to convert.
                    pass
            
            else:
            
                # Normal text.
                output.append(c)
        
        text_output = u"".join(output)
        
        # Return the generated item string.
        if text_output.strip() != u"":
        
            # Pad the string with a space if it is not complete whitespace.
            return text_output + u" "
        
        else:
        
            # Don't return pure whitespace; return an empty string instead.
            return u""
    
    def write_number(self, value):
    
        return u"%i" % value
    
    def write_alpha(self, value):
    
        if 0 < value < 27:
        
            return unicode(chr(96 + value))
        
        else:
        
            return u"?"
    
    def write_Alpha(self, value):
    
        return self.write_alpha(value).upper()
    
    def export_urls(self):
    
        for item in self.body:
        
            if isinstance(item, self.AnchorBegin):
            
                # Determine whether the referenced item is to be found
                # relative to the current page.
                
                if item.attributes[u"href"].find(u":") == -1:
                
                    # Assume it is a relative link.
                    
                    # Look for a suffix.
                    suffix_at = item.attributes[u"href"].rfind(u".")
                    
                    # Look for a hash symbol.
                    hash_at = item.attributes[u"href"].find(u"#")
                    
                    if hash_at == -1:
                    
                        # No named anchor was found.
                        url_part = item.attributes[u"href"] + u".html"
                        hash_part = u""
                        hash_at = len(item.attributes[u"href"])
                    
                    elif hash_at == 0:
                    
                        # Only a named anchor was found.
                        url_part = u""
                        hash_part = item.attributes[u"href"]
                    
                    else:
                    
                        url_part = item.attributes[u"href"][:hash_at] + u".html"
                        hash_part = item.attributes[u"href"][hash_at:]
                    
                    if (suffix_at == -1 or \
                        item.attributes[u"href"][suffix_at:hash_at] \
                            not in self.Image.image_formats):
                    
                        # The file either has no suffix or the suffix does
                        # not belong to that of a recognised image type.
                        
                        # Give the reference an ".html" suffix.
                        item.update_attributes( ((u"href", url_part + hash_part),) )
    
    def process_body(self):
    
        # Process the elements in the list.
        
        body = []
        
        indent = 0
        
        for element in self.body:
        
            # If this element supplies a process method then use it
            # to modify the elements previously added to the list.
            if hasattr(element, "process"):
            
                indent = element.process(body, indent)
            
            else:
            
                body.append(element)
        
        # Count the number of BlockBegin and BlockEnd elements used in the
        # document.
        div_begin = len(filter(lambda x: isinstance(x, self.BlockBegin), body))
        div_end = len(filter(lambda x: isinstance(x, self.BlockEnd), body))
        
        # If there are more beginning elements than ending elements then
        # add some ending elements.
        if div_begin > div_end:
        
            body = body + [self.BlockEnd(u"div")] * (div_begin - div_end)
        
        return u"".join(map(lambda element: element.text, body))
