#!/usr/bin/env python

"""
embedwikiplugin.py

KPart Wiki plugin for KHTML.

Copyright (C) 2003 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

__version__ = "0.40"

# Import the required Qt and KDE classes.

from qt import QHBox, QLabel, QMessageBox, QObject, QPoint, \
               QSplitter, QString, QStringList, QVBoxLayout, QWidget, \
               SIGNAL, PYSIGNAL

from kdecore import i18n, KAboutData, KConfig, KInstance, KLibFactory, \
                    KLibLoader, KShortcut, KTempFile, KURL
                    
from kparts import KParts, createReadWritePart
from kdeui import KAction, KActionMenu, KComboBox, KDialogBase, KLineEdit, \
                  KPushButton, KIntSpinBox, KStatusBar

# Provide a fallback option for when KTextEdit is not available. This has
# only been observed on one system.

try:

    from kdeui import KTextEdit
    TextEdit = KTextEdit

except ImportError:

    from qt import QTextEdit
    TextEdit = QTextEdit

from khtml import KHTMLPart, KHTMLView, DOM
from kio import KIO, KRun, KServiceTypeProfile

from kfile import KFileDialog

import glob, imp, os

if not os.__dict__.has_key("extsep"):

    os.extsep = "."

this_dir, this_file = os.path.split(__file__)
rc_file = "embedwikiplugin" + os.extsep + "rc"


# Define the default MIME types handled by this KPart.

MIMEtypes = (u"text/plain", u"text/x-readme")


# Debugging variable and functions --------------------------------------------

import sys, traceback

# Debugging flag: set to 1 to enable debugging.

DEBUG = 1

def debug_file():

    return sys.stderr

def debug(msg):

    if DEBUG == 0: return
    
    debug_file().write(msg+"\n")
    debug_file().flush()

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open("/tmp/pydebug.txt", "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": debug_file()}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    _instance = None
    
    def __init__(self, parent = None, name = None):
    
        KLibFactory.__init__(self, parent, name)
        
        self.name = "embedwikipart"
        self.version = __version__
        self.program_name = str(i18n("Embedded Wiki Part"))
        self.program_description = str(
            i18n("Create HTML pages using Wiki-style text markup.")
            )
        
        self.thanks = str(i18n(
            "Thanks to Jim Bublitz and Phil Thompson for providing advice " + \
            "on using PyKDE, PyQt and sip."
            ))
        
        self.authors = \
        [
            ( "David Boddie", str(i18n("Module author")),
              "david@boddie.org.uk", "http://www.boddie.org.uk/" )
        ]
        
        # The QStrings returned from the i18n function need to be
        # converted to strings for the following constructor.
        # The program name and program description strings must also be
        # translated and defined elsewhere rather than being translated
        # and converted to strings inline.
        
        self._about_data = KAboutData(
            self.name, self.program_name, self.version,
            self.program_description,
            KAboutData.License_GPL, "(c) 2003, David Boddie",
            self.thanks
            )
            
        # You _must_ set the license text if the License_Custom
        # value is used in the constructor.
        #self._about_data.setLicenseText(self.license)
        
        # It looks like all arguments are required for the addAuthor
        # method of the KAboutData class.
        
        for author, role, email, website in self.authors:
        
            self._about_data.addAuthor(
                author, role, email, website
                )
        
        if FactoryClass._instance is None:
        
            FactoryClass._instance = KInstance(self._about_data)
    
    def createObject(self, parent = None, name = None, class_name = "QObject",
                     args = None):
    
        if args is None:
        
            args = []
        
        plugin = EmbedWikiPart(parent, name)
        
        # Either of these appear to work:
        #self.emit(SIGNAL("objectCreated(KParts::Plugin)"), (plugin,))
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        
        return plugin
    
    create = createObject
    
    def instance(self):
    
        return FactoryClass._instance
    
    def aboutData(self):
    
        return self._about_data


# Define the plugin class, derived from the standard KParts.Plugin class.

class EmbedWikiPart(KParts.ReadWritePart):

    def __init__(self, parent = None, name = None):
    
        KParts.ReadWritePart.__init__(self, parent, name)
        
        # From the PartBase class:
        self.setInstance(FactoryClass._instance)
        
        # Set the XML file containing the XMLGUI definitions.
        self.setXMLFile( os.path.join(this_dir, rc_file) )
        
        #self.name = name
        #self.parent = parent
        
        actions = self.actionCollection()
        
        # For possible future expansion:
        #actions.setXMLFile( os.path.join(this_dir, rc_file) )
        
        self.edit_action = KAction(
            i18n("Revert to saved"), "embedwikirevert", KShortcut.null(),
            self.openFile, actions, "embedwikirevert"
            )
        
        self.save_action = KAction(
            i18n("&Save page"), "filesave", KShortcut.null(),
            self.slotSave, actions, "embedwikisave"
            )
        
        self.update_action = KAction(
            i18n("Update page"), "embedwikiupdate", KShortcut.null(),
            self.slotUpdate, actions, "embedwikiupdate"
            )
            
        self.viewhtml_action = KAction(
            i18n("View HTML"), "embedwikiviewhtml", KShortcut.null(),
            self.slotViewHTML, actions, "embedwikiviewhtml"
            )
            
        self.settings_action = KAction(
            i18n("Configure Wiki"), "embedwikisettings", KShortcut.null(),
            self.slotSettings, actions, "embedwikisettings"
            )
            
        self.savehtml_action = KAction(
            i18n("Save HTML"), "filesave", KShortcut.null(),
            self.slotSaveHTML, actions, "embedwikisavehtml"
            )
            
        # Create an extension object.
        self.m_extension = EmbedWikiBrowserExtension(self)
        
        # Load the configuration settings.
        self.load_config()
        
        # Create a part widget.
        self.part_widget = self.create_part_widget(parent)
        
        # From Part:
        self.setWidget(self.part_widget)
        
        # Set the part widget's focus policy.
        self.part_widget.setFocusPolicy(QWidget.ClickFocus)
        
        # Now that everything is set up, apply the configuration settings.
        self.apply_config()
        
        # Use an attribute to contain incoming data.
        self.file_data = []
    
    __init__ = trap(__init__)
    
    def load_config(self):
    
        self.part_config = WikiConfig("embedwikiplugin")
        
        self.part_config.read_config()
    
    def save_config(self):
    
        self.part_config = KConfig("embedwikiplugin", 0)
        
        self.part_config.write_config()
    
    def apply_config(self):
    
        """apply_config(self)
        
        Apply changes to the configuration of the KPart to the current
        display. This must only be called after the configuration object,
        main widget structure and HTML view have been set up.
        """
        
        if not (hasattr(self, "html") and hasattr(self, "part_config")):
        
            return
        
        # Allocate proportional amounts of space to each of the splitter's
        # child widgets.
        self.splitter.setSizes(
            [ self.part_config["Appearance"]["HTML space"].value,
              self.part_config["Appearance"]["Editor space"].value ]
            )
        
        # Find the name of the module to use to generate HTML from text.
        module_name = str(self.part_config["Wiki"]["Module"].value)
        
        # Load the module from a directory alongside this program.
        try:
        
            self.wiki_module = imp.load_source(
                "wiki",
                os.path.join(
                    this_dir + os.sep + "Modules",
                    module_name + os.extsep + "py"
                    )
                )
            
            # Create an object to convert text to HTML.
            self.htmlwriter = self.wiki_module.HTMLWriter()
        
        except (IOError, ImportError):
        
            self.warning("Configuration error", "Cannot find Wiki module.")
            
            # Create an object to convert text to HTML.
            self.htmlwriter = HTMLWriter()
        
        # Refresh the HTML display.
        if hasattr(self, "wiki_text"):
        
            self.htmlwriter.write(
                self.wiki_text, unicode(self.url().url())
                )
            
            wiki_html = self.htmlwriter.output()
            
            # Add the HTML to the HTML widget.
            self.html.setText(wiki_html, self.url())
    
    def create_part_widget(self, parent):
    
        # Create the widget to be used as the part widget.
        part_widget = QWidget(parent)
        
        layout = QVBoxLayout(part_widget)
        
        self.splitter = QSplitter(QSplitter.Vertical, part_widget)
        
        self.html = HTMLPart(
            self.splitter,
            "<html><body><em>Please wait...</em></body></html>"
            )
        
        self.html.view().reparent(self.splitter, QPoint(0, 0))
        
        # Connect the Python signal for selecting URLs from the HTMLPart
        # to a method in this instance.
        self.connect(self.html, PYSIGNAL("urlSelected"), self.slotURL)
        
        # Create an widget for text editing.
        editor = self.create_editor(self.splitter)
        
        # Create a status bar.
        self.status = KStatusBar(part_widget)
        
        # Connect the onURL signal from the HTMLPart widget to a method
        # in this instance.
        self.connect(
            self.html, SIGNAL("onURL(const QString &)"), self.slotOnURL
            )
        
        
        # Add the splitter and status bar widgets to the main part widget
        # but ensure that the splitter takes up most of the available
        # space.
        
        layout.addWidget(self.splitter, 1)
        layout.addWidget(self.status, 0)
        
        # Return the part widget.
        return part_widget
    
    def create_editor(self, parent):
    
        # Find the preferred component to handle this MIME type.
        offers = KServiceTypeProfile.offers(
            "text/plain", "KParts/ReadWritePart"
            )
        
        # Ensure that we are not recursively using this plugin by checking
        # each KServiceOffer instance to determine which refer to this KPart.
        offers = filter(
            lambda offer:
            str( offer.service().library() ).find("embedwikiplugin") == -1,
            offers
            )
        
        # Retrieve the part corresponding to the first available service.
        # Use an instance variable to store the editor part to prevent it
        # from going out of scope.
        
        libloader = KLibLoader.self()
        
        self.editor_part = None
        
        for service_offer in offers:
        
            # Note: A PyKDE specific function is used.
            service = service_offer.service()
            
            factory = libloader.factory(str(service.library()))
            
            args = QStringList()
            
            self.editor_part = createReadWritePart(
                service.library(), self.splitter, "editor",
                "KParts::ReadWritePart", args
                )
            
            if self.editor_part: break
        
        # Check for the absence of a suitable editor KPart.
        
        if not self.editor_part:
        
            # Use a TextEdit instance instead.
            
            self.editor_part = EditorWrapper(parent)
            return self.editor_part.widget()
        
        # Merge the XML file of the editor with that of the browser.
        editor_interface = self.editor_part.domDocument()
        
        # Merge the GUI found with the GUI of our part.
        
        #if xmlFile:
        #
        #    self.setXMLFile(xmlFile)
        
        # Find the part's widget and reparent it.
        widget = self.editor_part.widget()
        self.editor_part.widget().reparent(parent, QPoint(0, 0), 1)
        
        return self.editor_part.widget()
    
    def saveFile(self):
    
        # This method is a pure virtual method in KParts.ReadWritePart.
        
        # Try to write to the file.
        return self.editor_part.saveAs(self.url())
    
    def edit_warning(self):
    
        self.warning(
            "Cannot edit page",
            "You cannot edit anything except plain text with\n"
            "this plugin, sorry."
            )
    
    def warning(self, title, text):
    
        title = i18n( title )
        text = i18n( text )
        
        QMessageBox.warning( None, title, text )
    
    def openFile(self):
    
        if isinstance(self.editor_part, EditorWrapper):
        
            # Read the contents of the resource at the current URL.
            self.job = KIO.get(self.url(), 0, 1)
            
            # Define a list to contain the incoming data.
            self.file_data = []
            
            self.connect(
                self.job, SIGNAL("data(KIO::Job * , const QByteArray &)"),
                self.slotReadData
                )
            
            self.connect(
                self.job, SIGNAL("result(KIO::Job *)"), self.slotGetResult
                )
            
            self.connect(
                self.job, SIGNAL("mimetype(KIO::Job *, const QString &)"),
                self.slotMIMEType
                )
        
        else:
        
            # Use the editor part to load the text.
            self.connect(
                self.editor_part, SIGNAL("completed()"), self.slotReadEditor
                )
            
            self.editor_part.openURL(self.url())
        
        # Return a true value to indicate that we are using the normal
        # network transparency infrastructure.
        return 1
    
    def slotReadData(self, job, bytearray):
    
        if job != self.job or bytearray.data() is None:
        
            return
        
        # Use QString to determine the encoding from the text.
        udata = unicode(QString(bytearray))
        
        self.file_data.append(udata)
    
    def slotMIMEType(self, job, mimetype):
    
        # Store the MIME type for later use.
        self.mimetype = unicode(mimetype)
    
    def slotGetResult(self, job):
    
        if job != self.job:
        
            return
        
        elif job.error():
        
            self.edit_warning()
            return
        
        elif self.mimetype not in MIMEtypes:
        
            # Content of an incorrect type was returned.
            self.edit_warning()
            return
        
        # At this point, the loading process is complete.
        
        # Join all the pieces of data together and store it in an instance
        # variable.
        self.wiki_text = u"".join(self.file_data)
        
        # Convert the textual representation into a DOM representation of
        # the page.
        
        self.htmlwriter.write(
            self.wiki_text, unicode(self.url().url())
            )
        
        wiki_html = self.htmlwriter.output()
        
        if wiki_html is None:
        
            self.edit_warning()
            return
        
        # Add the text to the text widget.
        self.write_editor_text(wiki_text)
        
        # Add the HTML to the HTML widget.
        self.html.setText(wiki_html, self.url())
        
        # Set the URL in the browser's location bar.
        self.m_extension.setLocationBarURL(self.url().prettyURL())
        
        self.emit( SIGNAL("completed()"), () )
    
    def slotReadEditor(self):
    
        # Disconnect the completed signal from the editor part otherwise
        # we will end up here again once the next method is completed.
        self.disconnect(
            self.editor_part, SIGNAL("completed()"), self.slotReadEditor
            )
        
        # Retrieve the text from the editor and store it in a
        # temporary file.
        self.connect(
            self.editor_part, SIGNAL("completed()"), self.slotReadTemp
            )
        
        self.save_editor_text()
    
    def slotReadTemp(self):
    
        # Disconnect the completed signal from the editor part.
        self.disconnect(
            self.editor_part, SIGNAL("completed()"), self.slotReadTemp
            )
        
        # Read the text from the temporary file.
        self.wiki_text = self.read_temp_file()
        
        # Convert the textual representation into a DOM representation of
        # the page.
        
        self.htmlwriter.write(
            self.wiki_text, unicode(self.url().url())
            )
        
        wiki_html = self.htmlwriter.output()
        
        if wiki_html is None:
        
            self.edit_warning()
            return
        
        # Add the HTML to the HTML widget.
        self.html.setText(wiki_html, self.url())
        
        # Set the URL in the browser's location bar.
        self.m_extension.setLocationBarURL(self.url().prettyURL())
        
        self.emit( SIGNAL("completed()"), () )
    
    def save_editor_text(self):
    
        if isinstance(self.editor_part, EditorWrapper):
        
            return unicode(self.editor_part.widget().text())
        
        else:
        
            self.tempfile = KTempFile(QString(), os.extsep+"txt")
            temp_name = unicode(self.tempfile.name())
            
            self.editor_part.saveAs(KURL(u"file:"+temp_name))
        
    def read_temp_file(self):
    
        temp_name = unicode(self.tempfile.name())
        
        try:
        
            text = open(temp_name, "rb").read()
            self.tempfile.unlink()
        
        except IOError:
        
            return None
        
        # Use QString to determine the encoding from the text.
        utext = unicode(QString(text))
        
        return utext
    
    def write_editor_text(self, text):
    
        if isinstance(self.editor_part, EditorWrapper):
        
            self.editor_part.widget().setText(wiki_text)
            self.editor_part.m_url = self.url()
        
        else:
        
            #tempfile = KTempFile(None, os.extsep+"txt")
            #temp_name = unicode(tempfile.name())
            #
            #try:
            #
            #    text = open(temp_name, "wb").write(text.encode("utf_8"))
            #
            #except IOError:
            #
            #    return
            #
            #self.editor_part.openURL(KURL(u"file:"+temp_name))
            #tempfile.unlink()
            self.editor_part.openURL(self.url())
    
    def slotUpdate(self):
    
        # Read the text from the text widget, convert it, and write it
        # to the HTML widget. Store the text in an instance variable.
        
        # Retrieve the text from the editor and store it in a
        # temporary file.
        self.connect(
            self.editor_part, SIGNAL("completed()"), self.slotUpdateTemp
            )
        
        self.save_editor_text()
    
    def slotUpdateTemp(self):
    
        # Disconnect the completed signal from the editor part.
        self.disconnect(
            self.editor_part, SIGNAL("completed()"), self.slotUpdateTemp
            )
        
        # Read the text from the temporary file.
        self.wiki_text = self.read_temp_file()
        
        self.htmlwriter.write(
            self.wiki_text, unicode(self.url().url())
            )
        
        html = self.htmlwriter.output()
        
        if html is not None:
        
            self.html.setText(html, self.url())
        
        else:
        
            title = i18n( "Update page" )
            text = i18n( "Please open a page for editing." )
            
            QMessageBox.warning( None, title, text )
    
    def textChangedSlot(self):
    
        # We are not interested in the key pressed, only in whether the
        # editor text is modified.
        if self.editor_part.isModified():
        
            self.setModified()
    
    def slotSave(self):
    
        # Update the HTML view and save the contents of the text editor
        # widget.
        self.save()
        self.slotUpdate()
    
    def slotViewHTML(self):
    
        # Retrieve the text from the editor and store it in a
        # temporary file.
        self.connect(
            self.editor_part, SIGNAL("completed()"), self.slotViewHTMLTemp
            )
        
        self.save_editor_text()
    
    def slotViewHTMLTemp(self):
    
        # Disconnect the completed signal from the editor part.
        self.disconnect(
            self.editor_part, SIGNAL("completed()"), self.slotViewHTMLTemp
            )
        
        # Read the text from the temporary file.
        self.wiki_text = self.read_temp_file()
        
        self.tempfile = KTempFile(None, os.extsep+"html")
        temp_name = unicode(self.tempfile.name())
        
        try:
        
            self.htmlwriter.write(
                unicode(self.wiki_text),
                unicode(self.url().url())
                )
            
            open(temp_name, "wb").write( self.htmlwriter.output() )
            
            self.run = KRun.runURL(KURL(u"file:/"+temp_name), "text/plain", 1)
            
            if self.run == 0:
            
                raise IOError
        
        except IOError:
        
            self.warning("View HTML", "Could not write document.")
        
        #self.connect(self.run, SIGNAL("finished()"), self.slotViewFinished)
    
    def slotViewFinished(self):
    
        # Remove the temporary file.
        #self.tempfile.unlink()
        pass
    
    def slotSaveHTML(self):
    
        # Retrieve the text from the editor and store it in a
        # temporary file.
        self.connect(
            self.editor_part, SIGNAL("completed()"), self.slotSaveHTMLTemp
            )
        
        self.save_editor_text()
    
    def slotSaveHTMLTemp(self):
    
        # Disconnect the completed signal from the editor part.
        self.disconnect(
            self.editor_part, SIGNAL("completed()"), self.slotSaveHTMLTemp
            )
        
        # Find the destination for the HTML.
        filename = unicode(KFileDialog.getSaveFileName())
        
        # Read the text from the temporary file.
        self.wiki_text = self.read_temp_file()
        
        try:
        
            self.htmlwriter.write(
                unicode(self.wiki_text),
                unicode(self.url().url())
                )
            
            open(filename, "wb").write( self.htmlwriter.output(export = 1) )
        
        except IOError:
        
            self.warning("Save HTML", "Could not write document.")
    
    def slotURL(self, url):
    
        if KURL.isRelativeURL(url):
        
            # A relative URL of an appropriate type of file.
            
            if unicode(url)[:1] == u"#":
            
                # Notify the browser that we will be opening this file
                # ourselves.
                self.m_extension.openURLNotify()
                
                # Ask the HTML part to go to the anchor specified.
                self.html.gotoAnchor(url[1:])
                
                # Set the URL in the browser's location bar.
                self.m_extension.setLocationBarURL(
                    KURL( KURL(self.url()), url ).prettyURL()
                    )
                
                return
            
            else:
            
                # Join the current URL and the new URL.
                url = KURL(KURL(self.url()), url)
        
        else:
        
            url = KURL(url)
        
        # Check the MIME type of the resource referred to by the URL.
        self.job = KIO.mimetype(url, 0)
        
        self.connect(
            self.job, SIGNAL("result(KIO::Job *)"), self.slotURLGet
            )
    
    def slotURLGet(self, job):
    
        if job != self.job:
        
            return
        
        elif job.error():
        
            url = self.job.url()
            
            # If the URL corresponds to a local file then check for its
            # presence.
            
            if unicode(url.protocol()) == u"file":
            
                path = unicode(url.path())
                directory, file = os.path.split(path)
                
                if not os.path.exists(path):
                
                    # The file does not exist.
                    
                    # Create a new file at that location with some text from
                    # a template.
                    try:
                    
                        open(path, "w").write(u"="+file+u"=\n")
                    
                    except IOError:
                    
                        self.warning(
                            "Path error",
                            "Cannot use path to object specified."
                            )
                        return
            
            else:
            
                self.warning(
                    "Page creation error",
                    "Cannot create new remote pages."
                    )
                return
            
            # We will notify the browser that we will be opening this file
            # ourselves.
        
        elif unicode(self.job.mimetype()) in MIMEtypes:
        
            # We will notify the browser that we will be opening this file
            # ourselves.
            pass
        
        else:
        
            # Create a KURL object to use to open the new URL.
            
            # Ask the browser to open the page specified by the URL.
            self.m_extension.openURLRequest(self.job.url())
            
            # Open the page specified by the URL and inform the browser.
            #self.openURL(KURL(url))
            return
        
        # Notify the browser that we will be opening this file ourselves.
        self.m_extension.openURLNotify()
        
        # Open the URL in this part using the standard mechanism.
        self.openURL(self.job.url())
    
    def slotOnURL(self, url):
    
        # Show the URL in the status bar.
        self.status.message(url)
    
    def slotSettings(self):
    
        # Open a configuration dialogue window.
        if hasattr(self, "config_dialog"):
        
            del self.config_dialog
        
        self.config_dialog = WikiConfigDialog(self.part_config)
        self.config_dialog.show()
        
        self.connect(
            self.config_dialog, PYSIGNAL("configured"), self.apply_config
            )
    
    # Currently unused, but useful methods
    
    def find_children(self, obj, look_for):
    
        found = []
        
        if obj.inherits(look_for):
        
            found.append(obj)
        
        for child in obj.children():
        
            found = found + self.find_children(child, look_for)
        
        return found
    
    def find_part(self, obj, look_for):
    
        if obj.inherits(look_for):
        
            return obj
        
        elif obj.inherits("KParts::MainWindow"):
        
            part_obj = obj.partObject()
            
            if obj == part_obj: return None
            
            return self.find_part(obj, look_for)
        
        else:
        
            return None

class EmbedWikiBrowserExtension(KParts.BrowserExtension):

    def __init__(self, parent = None, name = None):
    
        KParts.BrowserExtension.__init__(self, parent, name)
    
    def setLocationBarURL(self, url):
    
        self.emit( SIGNAL("setLocationBarURL(QString &)"), (url,) )
    
    def openURLRequest(self, url, args = None):
    
        if args is None: args = KParts.URLArgs()
        
        self.emit(
            SIGNAL("openURLRequest(const KURL &, const KParts::URLArgs)"),
            (url, args)
            )
    
    def openURLNotify(self):
    
        self.emit( SIGNAL("openURLNotify()"), () )


# Create a helper class for displaying HTML and responding to events.

class HTMLPart(KHTMLPart):

    def __init__(self, parent, text = None, url = None):
    
        KHTMLPart.__init__(self, parent)
        
        if text is not None and url is not None:
        
            self.setText(text, url)
    
    def setText(self, text, url):
    
        self.begin(url)
        self.write(text)
        self.end()
    
    def text(self):
    
        # Starting at the root document element, convert relevant elements
        # into a simplified format.
        dom = self.document()
        
        root = dom.documentElement()
        
        return self.element_to_text(root)
    
    def element_to_text(self, node, indent = 0):
    
        wiki_text = []
        
        name = unicode(node.nodeName().string())
        value = unicode(node.nodeValue().string())
        
        attrs = node.attributes()
        attributes = {}
        
        for a in range(0, attrs.length()):
        
            item = attrs.item(a)
            attr_name = unicode(item.nodeName().string()).lower()
            attributes[attr_name] = unicode(item.nodeValue().string())
        
        if value == u"":
        
            wiki_text.append((u" "*indent) + u"<" + name.lower())
            
            if attributes != {}:
            
                wiki_text.append(u" ")
                
                wiki_text.append(
                    u" ".join(
                        map(
                            lambda x: x[0] + u"='" + x[1] + u'"',
                            attributes.items()
                            )
                        )
                    )
            
            wiki_text.append(u">")
            
            child = node.firstChild()
            
            while not child.isNull():
            
                wiki_text.append(self.element_to_text(child, indent + 1))
                
                child = child.nextSibling()
            
            wiki_text.append((u" "*indent) + u"</"+name.lower()+u">")
        
        else:
        
            wiki_text.append(value)
        
        return u"".join(wiki_text)
    
    def urlSelected(self, url, button, state, target, args):
    
        # Ideally we would emit a signal here. Instead, we use a Python
        # signal.
        self.emit( PYSIGNAL("urlSelected"), (url,))
        #self.run = KParts.BrowserRun(KURL(url), args, None, None, 1, 0)


# Create a class which mimics the KParts.ReadWritePart interface but which
# uses a standard editing widget to provide text editing functions.

class EditorWrapper(KParts.ReadWritePart):

    def __init__(self, parent):
    
        KParts.ReadWritePart.__init__(self, parent)
        
        self._widget = TextEdit(parent)
        
        # Connect the signal indicating changes in the text in the editor
        # to a slot in this instance.
        self.connect(
            self, SIGNAL("textChanged()"), self.textChangedSlot
            )
        
        self.modified = 0
    
    def textChangedSlot(self):
    
        self.modified = 1
    
    def isModified(self):
    
        return self.modified
    
    def widget(self):
    
        return self._widget
    
    def saveFile(self):
    
        # Unfortunately, the protected member "m_file" is not available
        # to us, so we must use the URL instead.
        if not hasattr(self, "url"):
        
            return 0
        
        # Split the URL into pieces.
        url_pieces = unicode(self.url().url()).split(u":")
        
        # The path is made up of all the pieces after the first piece.
        path = u":".join(url_pieces[1:])
        
        try:
        
            open(path, "wb").write(
                unicode(self._widget.text())
                )
        
        except IOError:
        
            return 0
        
        return 1
    
    def url(self):
    
        if hasattr(self, "m_url"):
        
            return self.m_url
        
        else:
        
            return KParts.ReadWritePart.url(self)

# A default text processing class for use when all else fails.

class HTMLWriter:

    head = \
    (
        u"<html>\n"
        u"<head>\n"
        u'  <meta http-equiv="Content-Type" '
        u'content="text/html; charset=utf-8">\n'
        u"  <title>%s</title>\n"
        u'  <style type="text/css">li.empty { list-style-type: none }</style>\n'
        u"</head>\n"
        u"\n"
        u"<body>\n"
        u"<pre>\n"
    )
    
    tail = \
    (
        u"</pre>\n"
        u"</body>\n"
        u"</html>\n"
    )
    
    def write(self, text, url):
    
        self.text = text
        self.url = url
    
    def output(self, export = 0):
    
        return u"".join(
            [ self.head % self.url, self.html_repr(self.text), self.tail ]
            )
    
    def html_repr(self, s):
    
        new = []
        
        for c in unicode(s):
        
            if c == u"&":
                new.append(u"&amp;")
            elif c == u"<":
                new.append(u"&lt;")
            elif c == u">":
                new.append(u"&gt;")
            else:
                new.append(c)
        
        return u"".join(new)



class FileChoice:

    def __init__(self, subpath, pattern, default):
    
        # Use the global directory variable to establish the full path
        # to the choice of files.
        
        # Join the path together and ensure that the standard
        # representation of the pattern string is adjusted to suit the
        # current platform.
        
        self.choice = os.path.join(
            this_dir, subpath, os.extsep.join(pattern.split("."))
            )
        
        self.default = default
        self.value = default
    
    def files(self, full_path = 0, with_suffix = 0):
    
        """files(self)
        
        Return a list of files (with or without their suffixes) which fit the
        pattern supplied at initialisation.
        """
        
        files = glob.glob(self.choice)
        
        if full_path == 0:
        
            files = map(lambda path: os.path.split(path)[1], files)
        
        if with_suffix == 1:
        
            return files
        
        else:
        
            # Strip the suffixes from each file and return the list.
            rfiles = []
                
            for file in files:
            
                pieces = file.split(os.extsep)
                
                if len(pieces) > 1:
                
                    rfiles.append(os.extsep.join(pieces[:-1]))
                
                else:
                
                    rfiles.append(file)
            
            return rfiles
    
    def set_value(self, value):
    
        if value in self.files():
        
            self.value = value


class IntChoice:

    def __init__(self, mini, maxi, step, default):
    
        self.mini = mini
        self.maxi = maxi
        self.step = step
        self.default = default
        self.value = default
    
    def set_value(self, value):
    
        if self.mini <= value <= self.maxi:
        
            self.value = value

class StringChoice:

    def __init__(self, choices, default):
    
        self.choices = choices
        self.value = default
        self.default = default
    
    def set_value(self, value):
    
        self.value = value

class WikiConfig(KConfig):

    # Define some settings to look for in the configuration file.
    
    _settings = \
    {
        "Appearance":
        {
            "HTML space":   IntChoice(1, 10, 1, 1),
            "Editor space": IntChoice(1, 10, 1, 1)
        },
        
        "Wiki":
        {
            "Module":       FileChoice("Modules", "*.py", "wiki")
        }
    }
    
    def __init__(self, name):
    
        KConfig.__init__(self, name)
        
        self._group = None
        
        self.defaults()
    
    def defaults(self):
    
        # Make a copy of the default values and write it to a new settings
        # dictionary.
        
        for group, entries in self._settings.items():
        
            for key, choice in entries.items():
            
                self._settings[group][key].set_value(choice.default)
    
    def __getitem__(self, key):
    
        return self._settings[key]
    
    def getitem(self, key):
    
        # Find the item in the settings dictionary according to the current
        # group in use.
        group = self._settings.get(self._group, None)
        
        if group is not None:
        
            return group.get(key, "")
        
        else:
        
            return ""
    
    def setGroup(self, name):
    
        KConfig.setGroup(self, name)
        
        self._group = name
    
    def readEntry(self, name, default):
    
        value = KConfig.readEntry(self, name, default)
        
        # Store the value in the list of properties for this group.
        self._settings[self._group][name].set_value(value)
        
        return value
    
    def readNumEntry(self, name, default):
    
        value = KConfig.readNumEntry(self, name, default)
        
        # Store the value in the list of properties for this group.
        self._settings[self._group][name].set_value(value)
        
        return value
    
    def writeEntry(self, name, value):
    
        KConfig.writeEntry(self, name, value)
        
        # Store the value in the list of properties for this group
        # only if the type of the value agrees with the type of the
        # entry already present.
        try:
        
            old_value = self._settings[self._group][name].value
        
        except KeyError:
        
            return
        
        if type(old_value) != type(value):
        
            return
        
        self._settings[self._group][name].set_value(value)
    
    def read_config(self):
    
        # Examine the settings dictionary and use the existing types
        # of entries in each group to determine the methods to use to
        # retrieve the configuration settings.
        
        for group, entries in self._settings.items():
        
            self.setGroup(group)
            
            for key, choice in entries.items():
            
                if isinstance(choice, IntChoice):
                
                    self.readNumEntry(key, choice.default)
                
                elif isinstance(choice, FileChoice):
                
                    # A choice of files means a choice of strings.
                    # We supply the default file as a default value.
                    self.readEntry(key, choice.default)
                
                else:
                
                    self.readEntry(key, choice.default)
    
    def write_config(self):
    
        # Examine the settings dictionary and write the configuration
        # settings based on the contents of each group dictionary.
        
        for group, entries in self._settings.items():
        
            self.setGroup(group)
            
            for key, choice in entries.items():
            
                KConfig.writeEntry(self, key, choice.value)
        
        self.sync()


class WikiConfigDialog(KDialogBase):

    def __init__(self, config, parent = None):
    
        KDialogBase.__init__(
            self, KDialogBase.Tabbed, i18n("Configure the Wiki"),
            KDialogBase.Help | KDialogBase.Default | KDialogBase.Apply | \
                KDialogBase.Ok | KDialogBase.Cancel,
            KDialogBase.Ok, parent, "WikiConfig", 0, 1
            )
        
        self.config = config
        
        self.changed = 0
        
        self.pages = []
        
        self.create_pages()
    
    def create_pages(self):
    
        # Remove all current pages.
        for page, group, widgets in self.pages:
        
            page.reparent( None, QPoint(0, 0) )
        
        self.pages = []
        
        for group, entries in self.config._settings.items():
        
            page = self.addPage( i18n(str(group)) )
            
            widgets = self.create_page(page, entries)
            
            self.pages.append( (page, group, widgets) )
    
    def create_page(self, frame, entries):
    
        widgets = []
        
        # Create a layout widget for this frame.
        layout = QVBoxLayout(frame)
        
        for key, choice in entries.items():
        
            hbox = QHBox(frame)
            
            # Add a label for this item.
            QLabel(i18n(key), hbox)
            
            if isinstance(choice, IntChoice):
            
                # Create a spin box for the numbers.
                spinbox = KIntSpinBox(
                    choice.mini, choice.maxi, choice.step, choice.value,
                    10, hbox
                    )
                
                spinbox.connect(
                    spinbox, SIGNAL("valueChanged(int)"), self.slotChanged
                    )
                
                widgets.append( (key, spinbox, "value") )
            
            elif isinstance(choice, FileChoice):
            
                # Create a combo box to display the choices.
                combo_box = KComboBox(hbox)
                
                # Fill the combo box with the choices available.
                files = choice.files()
                combo_box.insertStrList(files)
                
                try:
                
                    index = files.index(choice.value)
                
                except ValueError:
                
                    index = 0
                
                combo_box.setCurrentItem(index)
                
                combo_box.connect(
                    combo_box, SIGNAL("textChanged(const QString &)"),
                    self.slotChanged
                    )
                
                widgets.append( (key, combo_box, "currentText") )
            
            else:
            
                # Add a line editor for strings.
                line_edit = KLineEdit(choice.value, hbox)
                
                line_edit.connect(
                    line_edit, SIGNAL("textChanged(const QString &)"),
                    self.slotChanged
                    )
                
                widgets.append( (key, line_edit, "text") )
                
            layout.addWidget(hbox)
        
        return widgets
    
    def read_pages(self):
    
        """read_pages(self)
        
        Read the contents of the pages into the configuration object
        but do not write the settings to the configuration file.
        """
        
        for page, group, widgets in self.pages:
        
            for name, widget, method_name in widgets:
            
                self.config[group][name].set_value(
                    getattr(widget, method_name)()
                    )
    
    def load(self):
    
        # Use the config object to load the settings again.
        self.config.read_config()
        
        self.create_pages()
    
    def defaults(self):
    
        # Use the config object to restore the default settings.
        self.config.defaults()
        
        self.create_pages()
    
    def slotOk(self):
    
        # Read the current values.
        self.read_pages()
        
        # Use the config object to save the settings.
        self.config.write_config()
        
        self.changed = 0
        
        self.emit( PYSIGNAL("configured"), () )
        
        self.close()
    
    def slotApply(self):
    
        # Read the current values.
        self.read_pages()
        
        self.emit( PYSIGNAL("configured"), () )
    
    def slotCancel(self):
    
        if self.changed == 1:
        
            answer = QMessageBox.warning(
                None, i18n("Embedded Wiki configuration"),
                i18n(
                    "The current configuration has not been saved.\n"
                    "Do you wish to save these settings?"
                    ),
                i18n("Yes"), i18n("No")
                )
            
            if answer == 0:
            
                # Proceed as if the user had clicked "OK" instead.
                self.slotOk()
            
            else:
            
                # Pretend that no changes had been made and call the
                # base class's slotCancel method.
                self.changed = 0
                KDialogBase.slotCancel(self)
        
        else:
        
            # Call the base class's method to deal with cancellation.
            KDialogBase.slotCancel(self)
    
    def slotDefault(self):
    
        self.config.defaults()
        
        self.create_pages()
    
    def slotChanged(self):
    
        self.changed = 1
