#!/usr/bin/env python

"""
pyspritepart.py

A KPart for displaying the contents of Acorn Spritefiles.

Copyright (C) 2003-2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# Import the required Qt and KDE classes.

import sip
from qt import QByteArray, QFrame, QGridLayout, QLabel, QObject, QPixmap, \
               QScrollView, QString, Qt, QVBox, QWidget, SIGNAL, QLayout

from kdecore import i18n, KAboutData, KInstance, KLibFactory, KShortcut, KURL
from kparts import KParts
from kdeui import KAction
from kio import KIO

# The os module is required for local path manipulation.

import os

# Import the spritefile module. If accelerated routines are unavailable then
# the module will still function, but it will be slower than usual.

import spritefile

# Try to import the C version of the StringIO class but fall back to the
# Python version if necessary.

try:

    from cStringIO import StringIO

except ImportError:

    from StringIO import StringIO


try:

    import Image
    Image.preinit()
    no_Image = 0

except ImportError:

    no_Image = 1



# Debugging variable and functions --------------------------------------------

import sys, traceback

# Debugging flag: set to 1 to enable debugging.

DEBUG = 0

def debug_file():

    return sys.stderr

def debug(msg):

    if DEBUG == 0: return
    
    debug_file().write(msg+"\n")
    debug_file().flush()

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open("/tmp/pydebug.txt", "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": debug_file()}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    s_instance = None
    
    def __init__(self, parent = None, name = None):
    
        #print "FactoryClass.__init__%s" % repr((parent, name))
        
        KLibFactory.__init__(self, parent, name)
        
        # Define some author information.
        
        self.name = "examplepart"
        self.version = "0.10"
        self.program_name = str(i18n("Example Part"))
        self.program_description = str(
            i18n("Prototype KPart written partly in Python.")
            )
        
        self.thanks = str(i18n(
            "Thanks to Jim Bublitz and Phil Thompson for providing advice " + \
            "on using PyKDE, PyQt and sip."
            ))
        
        self.authors = \
        [
            ( "David Boddie", str(i18n("Module author")),
              "david@boddie.org.uk", "http://www.boddie.org.uk/" )
        ]
        
        # The QStrings returned from the i18n function need to be
        # converted to strings for the following constructor.
        # The program name and program description strings must also be
        # translated and defined elsewhere rather than being translated
        # and converted to strings inline.
        
        self._about_data = KAboutData(
            self.name, self.program_name, self.version,
            self.program_description,
            KAboutData.License_GPL, "(c) 2003, David Boddie",
            self.thanks
            )
            
        # You _must_ set the license text if the License_Custom
        # value is used in the constructor.
        #self._about_data.setLicenseText(self.license)
        
        # It looks like all arguments are required for the addAuthor
        # method of the KAboutData class.
        
        for author, role, email, website in self.authors:
        
            self._about_data.addAuthor(
                author, role, email, website
                )
        
        if FactoryClass.s_instance is None:
        
            FactoryClass.s_instance = KInstance(self._about_data)
    
    def createObject(self, parent = None, name = None, class_name = "QObject",
                     args = None):
    
        if args is None:
        
            args = []
        
        plugin = ExamplePart(parent, name)
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        
        return plugin
    
    create = createObject
    
    def instance(self):
    
        return FactoryClass.s_instance
    
    def aboutData(self):
    
        return self._about_data


# Define a class, derived from the KParts.ReadOnlyPart class.

class ExamplePart(KParts.ReadOnlyPart):

    #m_extension = None
    
    def __init__(self, parent = None, name = None):
    
        KParts.ReadOnlyPart.__init__(self, parent, name)
        
        # From the PartBase class:
        self.setInstance(FactoryClass.s_instance)
        
        # Create an extension object.
        self.m_extension = ExampleBrowserExtension(self)
        
        # Create a widget and show it.
        self.part_widget = PartContainer(parent)
        
        # Create the contents of the part widget.
        self.contents = ExampleWidget(self.part_widget)
        self.part_widget.addChild(self.contents)
        
        # From Part:
        # Set the widget in the part.
        self.setWidget(self.part_widget)
        
        # Use an attribute to contain incoming data.
        self.file_data = []
    
    __init__ = trap(__init__)
    
    def openFile(self):
    
        # The value returned by self.url() is only defined if we implement
        # this pure virtual method.
        
        # Start a KIO.TransferJob to fetch the contents at the URL given.
        # The second boolean value indicates that we want to display a
        # status window.
        
        self.job = KIO.get(self.url(), 0, 1)
        
        #QObject.connect(self.job, SIGNAL("connected(KIO::Job *)"), self.connected)
        self.connect(self.job, SIGNAL("data(KIO::Job * , const QByteArray &)"), self.read_data)
        #self.connect(self.job, SIGNAL("canceled(KIO::Job *)"), self.slotCancel)
        self.connect(self.job, SIGNAL("result(KIO::Job *)"), self.slotResult)
        
        # Reset the attribute which contains incoming data.
        self.file_data = []
        
        # Send a signal indicating that a job has begun.
        self.emit(SIGNAL("started(KIO::Job *)"), (self.job,))
        
        # Return a true value to indicate that we are using the normal
        # network transparency infrastructure.
        return 1
    
    def read_data(self, job, bytearray):
    
        if bytearray.data() is None:
        
            return
        
        self.file_data.append(bytearray.data())
    
    #def slotCancel(self, job):
    #
    #    print "### Canceled: " + repr(job)
    
    def slotResult(self, job):
    
        #print "### Result: " + repr(job)
        if job.error():
        
            job.showErrorDialog()
            return
        
        # At this point, the loading process is complete.
        
        # Join all the pieces of data together.
        data = "".join(self.file_data)
        
        #print "%s is %i bytes in length." % (str(self.url().url()), len(data))
        
        # Create a widget and show it.
        
        self.part_widget.removeChild(self.contents)
        self.contents.destroy()
        
        self.contents = QFrame(self.part_widget)
        
        # Send a signal indicating that processing is complete.
        # Emit the completed signal.
        #print "###### Emit the completed signal."
        self.emit(SIGNAL("completed()"), ())
        
        self.part_widget.addChild(self.contents)
        
        # Read and display the sprites.
        sprites = self.read_sprites(data, self.contents)
        
        self.contents.show()
        
        #self.resize(self.parent.width(), self.parent.height())
        self.part_widget.updateContents()
    
    def read_sprites(self, data, frame):
    
        # Put the data in a StringIO object.
        string_io = StringIO(data)
        
        try:
        
            # Use the StringIO object to pass the data to the spritefile object.
            sprites = spritefile.spritefile(string_io)
        
        except spritefile.spritefile_error:
        
            return
        
        # Determine how many sprites there are.
        number = len(sprites.sprites.keys())
        
        columns = 5
        
        # Allocate enough rows for all the sprites, plus two more for padding.
        rows = int(number / columns)
        
        if number % columns > 0:
        
            rows = rows + 1
        
        layout = QGridLayout(frame, rows + 2, columns + 2, 32, 32)
        
        layout.setRowStretch(0,        1)
        layout.setRowStretch(rows + 1, 1)
        
        layout.setColStretch(0,           1)
        layout.setColStretch(columns + 1, 1)
        
        for row in range(1, rows + 1):
        
            layout.setRowStretch(row, 1)
        
        for column in range(1, columns + 1):
        
            layout.setColStretch(column, 1)
        
        # For each sprite, add a cell containing the image and another
        # with information.
        
        column = 0
        row = 1
        
        for name, details in sprites.sprites.items():
        
            vbox = QVBox(frame)
            
            if no_Image:
                pixmap = self.convert(details)
            else:
                pixmap = self.sprite_to_pixmap(details)
            
            label = QLabel(vbox)
            label.setPixmap(pixmap)
            label.setAlignment(QLabel.AlignCenter)
            
            caption = QLabel(vbox)
            caption.setText(QString(name))
            caption.setAlignment(QLabel.AlignCenter)
            
            layout.addWidget(vbox, row, column + 1)
            
            column = column + 1
            
            if column == columns:
            
                column = 0
                row = row + 1
    
    def convert(self, sprite):
    
        # Convert the data to a PPM image.
        
        width, height = sprite['width'], sprite['height']
        
        ppm_lines = []
        
        # Support rescaling for sprites with horizontal and vertical
        # resolutions that are not equal. The dots per inch in one direction
        # must be an integer multiple of the other for this to work.
        
        if sprite['dpi x'] > sprite['dpi y']:
        
            xfactor = 1
            yfactor = int(sprite['dpi x'] / sprite['dpi y'])
            ppm_lines.append(
                "P3 %i %i 255" % (width, yfactor * height)
                )
        
        elif sprite['dpi y'] > sprite['dpi x']:
        
            xfactor = int(sprite['dpi y'] / sprite['dpi x'])
            yfactor = 1
            ppm_lines.append(
                "P3 %i %i 255" % (xfactor * width, height)
                )
        
        else:
        
            xfactor = yfactor = 1
            ppm_lines.append("P3 %i %i 255" % (width, height))
        
        ptr = 0
        image = sprite['image']
        
        if sprite['mode'] == "RGB":
        
            for j in range(0, height):
            
                line = []
                
                for i in range(0, width):
                
                    line = line + [
                        "%i %i %i " % (
                            ord(image[ptr]), ord(image[ptr+1]), ord(image[ptr+2])
                            )
                        ] * xfactor
                    
                    ptr = ptr + 3
                
                ppm_lines = ppm_lines + ["".join(line)] * yfactor
        
        elif sprite['mode'] == "RGBA":
        
            for j in range(0, height):
            
                line = []
                
                for i in range(0, width):
                
                    line = line + [
                        "%i %i %i " % (
                            ord(image[ptr]), ord(image[ptr+1]), ord(image[ptr+2])
                            )
                        ] * xfactor
                    
                    ptr = ptr + 4
                
                ppm_lines = ppm_lines + ["".join(line)] * yfactor
        
        else:
        
            return QPixmap()
        
        ppm = QByteArray("\n".join(ppm_lines))
        length = len(ppm.data())
        
        # Create a QPixmap from the image data.
        pixmap = QPixmap()
        pixmap.loadFromData(ppm, "PPM", 0)
        
        return pixmap
    
    def sprite_to_pixmap(self, sprite):
    
        try:
        
            image = Image.fromstring(
                sprite['mode'], (sprite['width'], sprite['height']),
                sprite['image']
                )
        
        except KeyError:
        
            # Use the convert method instead.
            return self.convert(sprite)
        
        if sprite["dpi x"] != sprite["dpi y"]:
        
            # If the horizontal and vertical resolutions of the sprite are
            # not equal then rescale the image.
            if sprite["dpi x"] > sprite["dpi y"]:
            
                # There were fewer pixels per vertical unit of length so
                # increase the number of rows accordingly. This does not
                # try to find a new height which is an integer number of
                # pixels.
                yscale = float(sprite["dpi x"]) / sprite["dpi y"]
                image = image.resize( (image.size[0], image.size[1] * yscale) )
            
            else:
            
                # There were fewer pixels per vertical unit of length so
                # increase the number of rows accordingly. This does not
                # try to find a new height which is an integer number of
                # pixels.
                xscale = float(sprite["dpi x"]) / sprite["dpi y"]
                image = image.resize( (image.size[0] * xscale, image.size[1]) )
        
        # Put the data in a StringIO object.
        string_io = StringIO()
        
        for format in "PNG", "PPM", "TIFF", "BMP":
        
            if format in Image.SAVE.keys():
            
                string_io.seek(0)
                image.save(string_io, format = format)
                break
        
        if string_io.tell() == 0:
        
            # Use the convert method instead.
            return self.convert(sprite)
        
        string_io.seek(0)
        
        pixmap = QPixmap()
        pixmap.loadFromData(string_io.read())
        
        # Close the StringIO object.
        string_io.close()
        
        return pixmap


class PartContainer(QScrollView):

    def __init__(self, parent):
    
        QScrollView.__init__(self, parent)
        
        self.parent = parent
        
        self.setFocusPolicy(QWidget.ClickFocus)
        self.resize(parent.width(), parent.height())


class ExampleWidget(QWidget):

    def __init__(self, parent):
    
        QWidget.__init__(self, parent)
        
        self.parent = parent
        
        # Record the path to this module. We could obtain this from the general
        # launcher, since it has to locate this module anyway.
        
        self.module_path = __file__
        
        # Find a pixmap in the directory where this module resides.
        appdir, appmodule = os.path.split(self.module_path)
        
        pixmap_file = os.path.join(appdir, "image" + os.extsep + "png")
        
        if os.path.exists(pixmap_file):
        
            self.pixmap = QPixmap(pixmap_file)
        
        else:
        
            self.pixmap = None
        
        # Construct the user interface.
        self.construct_UI()
    
    def __del__(self):
    
        pass
    
    def construct_UI(self):
    
        if self.pixmap is not None:
        
            self.setBackgroundMode(self.FixedPixmap)
            self.setPaletteBackgroundPixmap(self.pixmap)
        
        else:
        
            self.setBackgroundMode(self.FixedColor)
            self.setPaletteBackgroundColor(Qt.green)
        
        self.resize(self.parent.width(), self.parent.height())


class ExampleBrowserExtension(KParts.BrowserExtension):

    def __init__(self, parent = None, name = None):
    
        KParts.BrowserExtension.__init__(self, parent, name)
