#!/usr/bin/env python

"""
restpart.py

reStructuredText Viewer KPart

Copyright (C) 2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

__version__ = "0.10"

# Import the required Qt and KDE classes.

from qt import QMessageBox, QObject, QPoint, QString, QVBoxLayout, QWidget, \
               SIGNAL, PYSIGNAL

from kdecore import i18n, KAboutData, KInstance, KLibFactory, \
                    KShortcut, KURL
from kfile import KFileDialog
from khtml import KHTMLPart
from kio import KIO
from kparts import KParts
from kdeui import KAction, KActionMenu, KComboBox

import glob, imp, os

import reST_writer

if not os.__dict__.has_key("extsep"):

    os.extsep = "."

this_dir, this_file = os.path.split(__file__)
rc_file = "restpart" + os.extsep + "rc"


# Define the default MIME types handled by this KPart.

MIMEtypes = (u"text/plain", u"text/x-readme")


# Debugging variable and functions --------------------------------------------

import sys, traceback

# Debugging flag: set to 1 to enable debugging.

DEBUG = False

def debug_file():

    return sys.stderr

def debug(msg):

    if DEBUG == 0: return
    
    debug_file().write(msg+"\n")
    debug_file().flush()

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open("/tmp/pydebug.txt", "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": debug_file()}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    _instance = None
    
    def __init__(self, parent = None, name = None):
    
        KLibFactory.__init__(self, parent, name)
        
        self.name = "reSTpart"
        self.version = __version__
        self.program_name = str(i18n("reStructuredText Viewer KPart"))
        self.program_description = str(
            i18n("View reStructuredText documents as HTML.")
            )
        
        self.thanks = str(i18n(
            "Thanks to Jim Bublitz and Phil Thompson for providing advice " + \
            "on using PyKDE, PyQt and SIP."
            ))
        
        self.authors = \
        [
            ( "David Boddie", str(i18n("Module author")),
              "david@boddie.org.uk", "http://www.boddie.org.uk/" )
        ]
        
        # The QStrings returned from the i18n function need to be
        # converted to strings for the following constructor.
        # The program name and program description strings must also be
        # translated and defined elsewhere rather than being translated
        # and converted to strings inline.
        
        self._about_data = KAboutData(
            self.name, self.program_name, self.version,
            self.program_description,
            KAboutData.License_GPL, "(c) 2005, David Boddie",
            self.thanks
            )
            
        # It looks like all arguments are required for the addAuthor
        # method of the KAboutData class.
        
        for author, role, email, website in self.authors:
        
            self._about_data.addAuthor(
                author, role, email, website
                )
        
        if FactoryClass._instance is None:
        
            FactoryClass._instance = KInstance(self._about_data)
    
    def createObject(self, parent = None, name = None, class_name = "QObject",
                     args = None):
    
        if args is None:
        
            args = []
        
        plugin = reSTPart(parent, name)
        
        # Either of these appear to work:
        #self.emit(SIGNAL("objectCreated(KParts::Plugin)"), (plugin,))
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        
        return plugin
    
    create = createObject
    
    def instance(self):
    
        return FactoryClass._instance
    
    def aboutData(self):
    
        return self._about_data


# Define the plugin class, derived from the standard KParts.Plugin class.

class reSTPart(KParts.ReadWritePart):

    def __init__(self, parent = None, name = None):
    
        KParts.ReadWritePart.__init__(self, parent, name)
        
        # From the PartBase class:
        self.setInstance(FactoryClass._instance)
        
        # Set the XML file containing the XMLGUI definitions.
        self.setXMLFile( os.path.join(this_dir, rc_file) )
        
        actions = self.actionCollection()
        
        # For possible future expansion:
        #actions.setXMLFile( os.path.join(this_dir, rc_file) )
        
        self.save_action = KAction(
            i18n("&Save page"), "filesave", KShortcut.null(),
            self.slotSave, actions, "reSTsave"
            )
        
        self.savehtml_action = KAction(
            i18n("Save HTML"), "filesave", KShortcut.null(),
            self.slotSaveHTML, actions, "reSTsavehtml"
            )
            
        # Create an extension object.
        self.m_extension = reSTPartBrowserExtension(self)
        
        # Create a part widget.
        self.part_widget = self.create_part_widget(parent)
        
        # From Part:
        self.setWidget(self.part_widget)
        
        # Set the part widget's focus policy.
        self.part_widget.setFocusPolicy(QWidget.ClickFocus)
        
        # Create an object to convert text to HTML.
        self.htmlwriter = reST_writer.HTMLWriter()
        
        # Use an attribute to contain incoming data.
        self.file_data = []
    
    __init__ = trap(__init__)
    
    def create_part_widget(self, parent):
    
        # Create the widget to be used as the part widget.
        part_widget = QWidget(parent)
        
        layout = QVBoxLayout(part_widget)
        
        self.html = HTMLPart(
            part_widget,
            "<html><body><em>Please wait...</em></body></html>"
            )
        
        self.html.view().reparent(part_widget, QPoint(0, 0))
        
        # Connect the Python signal for selecting URLs from the HTMLPart
        # to a method in this instance.
        self.connect(self.html, PYSIGNAL("urlSelected"), self.slotURL)
        
        # Connect the onURL signal from the HTMLPart widget to a method
        # in a status bar extension.
        self.status_bar_extension = KParts.StatusBarExtension(self)
        
        self.connect(
            self.html, SIGNAL("onURL(const QString &)"), self.setStatusBar
            )
        
        
        # Add the splitter and status bar widgets to the main part widget
        # but ensure that the splitter takes up most of the available
        # space.
        
        layout.addWidget(self.html.view())
        
        # Return the part widget.
        return part_widget
    
    def display_warning(self):
    
        self.warning(
            "Cannot display document",
            "Only reStructuredText documents can be viewed with\n"
            "this plugin."
            )
    
    def warning(self, title, text):
    
        title = i18n( title )
        text = i18n( text )
        
        QMessageBox.warning( None, title, text )
    
    def openFile(self):
    
        # Read the contents of the resource at the current URL.
        self.job = KIO.get(self.url(), 0, 1)
        
        # Define a list to contain the incoming data.
        self.file_data = []
        
        self.connect(
            self.job, SIGNAL("data(KIO::Job * , const QByteArray &)"),
            self.slotReadData
            )
        
        self.connect(
            self.job, SIGNAL("result(KIO::Job *)"), self.slotGetResult
            )
        
        self.connect(
            self.job, SIGNAL("mimetype(KIO::Job *, const QString &)"),
            self.slotMIMEType
            )
        
        # Return a true value to indicate that we are using the normal
        # network transparency infrastructure.
        return True
    
    def slotReadData(self, job, bytearray):
    
        if job != self.job or bytearray.data() is None:
        
            return
        
        # Use QString to determine the encoding from the text.
        udata = unicode(QString(bytearray))
        
        self.file_data.append(udata)
    
    def slotMIMEType(self, job, mimetype):
    
        # Store the MIME type for later use.
        self.mimetype = unicode(mimetype)
    
    def slotGetResult(self, job):
    
        if job != self.job:
        
            return
        
        elif job.error():
        
            self.display_warning()
            return
        
        elif self.mimetype not in MIMEtypes:
        
            # Content of an incorrect type was returned.
            self.display_warning()
            return
        
        # At this point, the loading process is complete.
        
        # Join all the pieces of data together and store it in an instance
        # variable.
        self.wiki_text = u"".join(self.file_data)
        
        # Convert the textual representation into a DOM representation of
        # the page.
        
        self.htmlwriter.write(
            self.wiki_text, unicode(self.url().url())
            )
        
        wiki_html = self.htmlwriter.output()
        
        if wiki_html is None:
        
            self.edit_warning()
            return
        
        # Add the HTML to the HTML widget.
        self.html.setText(wiki_html, self.url())
        
        # Set the URL in the browser's location bar.
        self.m_extension.setLocationBarURL(self.url().prettyURL())
        
        self.emit( SIGNAL("completed()"), () )
    
    def saveFile(self):
    
        # This method is a pure virtual method in KParts.ReadWritePart.
        
        # Try to write to the file.
        #return self.editor_part.saveAs(self.url())
        pass
    
    def slotSave(self):
    
        # Update the HTML view and save the contents of the text editor
        # widget.
        self.save()
    
    def slotSaveHTML(self):
    
        # Find the destination for the HTML.
        filename = unicode(KFileDialog.getSaveFileName())
        
        try:
        
            self.htmlwriter.write(
                unicode(self.wiki_text),
                unicode(self.url().url())
                )
            
            open(filename, "wb").write( self.htmlwriter.output(export = 1) )
        
        except IOError:
        
            self.warning("Save HTML", "Could not write document.")
    
    def slotURL(self, url):
    
        if KURL.isRelativeURL(url):
        
            # A relative URL of an appropriate type of file.
            
            if unicode(url)[:1] == u"#":
            
                # Notify the browser that we will be opening this file
                # ourselves.
                self.m_extension.openURLNotify()
                
                # Ask the HTML part to go to the anchor specified.
                self.html.gotoAnchor(url[1:])
                
                # Set the URL in the browser's location bar.
                self.m_extension.setLocationBarURL(
                    KURL( KURL(self.url()), url ).prettyURL()
                    )
                
                return
            
            else:
            
                # Join the current URL and the new URL.
                url = KURL(KURL(self.url()), url)
        
        else:
        
            url = KURL(url)
        
        # Check the MIME type of the resource referred to by the URL.
        self.job = KIO.mimetype(url, 0)
        
        self.connect(
            self.job, SIGNAL("result(KIO::Job *)"), self.slotURLGet
            )
    
    def slotURLGet(self, job):
    
        if job != self.job:
        
            return
        
        elif job.error():
        
            url = self.job.url()
            
            # If the URL corresponds to a local file then check for its
            # presence.
            
            if unicode(url.protocol()) == u"file":
            
                path = unicode(url.path())
                directory, file = os.path.split(path)
                
                if not os.path.exists(path):
                
                    # The file does not exist.
                    
                    # Create a new file at that location with some text from
                    # a template.
                    try:
                    
                        open(path, "w").write(u"="+file+u"=\n")
                    
                    except IOError:
                    
                        self.warning(
                            "Path error",
                            "Cannot use path to object specified."
                            )
                        return
            
            else:
            
                self.warning(
                    "Page creation error",
                    "Cannot create new remote pages."
                    )
                return
            
            # We will notify the browser that we will be opening this file
            # ourselves.
        
        elif unicode(self.job.mimetype()) in MIMEtypes:
        
            # We will notify the browser that we will be opening this file
            # ourselves.
            pass
        
        else:
        
            # Create a KURL object to use to open the new URL.
            
            # Ask the browser to open the page specified by the URL.
            self.m_extension.openURLRequest(self.job.url())
            
            # Open the page specified by the URL and inform the browser.
            #self.openURL(KURL(url))
            return
        
        # Notify the browser that we will be opening this file ourselves.
        self.m_extension.openURLNotify()
        
        # Open the URL in this part using the standard mechanism.
        self.openURL(self.job.url())
    
    def setStatusBar(self, text):
    
        self.status_bar_extension.statusBar().message(text)


class reSTPartBrowserExtension(KParts.BrowserExtension):

    def __init__(self, parent = None, name = None):
    
        KParts.BrowserExtension.__init__(self, parent, name)
    
    def setLocationBarURL(self, url):
    
        self.emit( SIGNAL("setLocationBarURL(QString &)"), (url,) )
    
    def openURLRequest(self, url, args = None):
    
        if args is None: args = KParts.URLArgs()
        
        self.emit(
            SIGNAL("openURLRequest(const KURL &, const KParts::URLArgs)"),
            (url, args)
            )
    
    def openURLNotify(self):
    
        self.emit( SIGNAL("openURLNotify()"), () )


# Create a helper class for displaying HTML and responding to events.

class HTMLPart(KHTMLPart):

    def __init__(self, parent, text = None, url = None):
    
        KHTMLPart.__init__(self, parent)
        
        if text is not None and url is not None:
        
            self.setText(text, url)
    
    def setText(self, text, url):
    
        self.begin(url)
        self.write(text)
        self.end()
    
    def urlSelected(self, url, button, state, target, args):
    
        # Ideally we would emit a signal here. Instead, we use a Python
        # signal.
        self.emit( PYSIGNAL("urlSelected"), (url,))
        #self.run = KParts.BrowserRun(KURL(url), args, None, None, 1, 0)
